/**CFile*******************************************************************
  PackageName [bdd]
  Synopsis    [Package 'bdd' enable symbolic computations by representing
               Boolean functions with ROBDDs.]

  FileName    [bddMain.c]
  Revision    [$Revision: 133 $]
  Date        [$Date: 2016-07-23 23:21:30 +0200 (sob, 23 jul 2016) $]
  Authors     [Robert Meolic (robert.meolic@um.si)]
  Description [File bddMain.c contains main functions for
               representation and manipulation of boolean functions
               with ROBDDs.]
  SeeAlso     [bdd.h, bddInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 2003, 2016
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include <biddy.h>
#include "bddInt.h"

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

/* ALL SIZES ARE unsigned int                                      */
/* Size of node and cache tables must be 2^N-1                     */
/* e.g. 2^10-1 = 1023, 2^16-1 = 65535, 2^20-1 = 1048575            */
/*      2^24-1 = 16777215                                          */

#define TINY_TABLE   1023
#define SMALL_TABLE  65535
#define BIG_TABLE    1048575
#define HUGE_TABLE   16777215

/*-----------------------------------------------------------------------*/
/* Macro definitions                                                     */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Type declarations                                                     */
/*-----------------------------------------------------------------------*/

typedef void (*Bdd_GCFunction)(Biddy_Manager);

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/* EXPORTED VARIABLES */

int bdd_status = 0;                             /* initialization status */

Bdd_Edge bdd_termTrue = {NULL,FALSE};                   /* terminal node */
Bdd_Edge bdd_termFalse = {NULL,FALSE};                  /* terminal node */
Bdd_Edge bdd_termNull = {NULL,FALSE};                       /* null edge */

/* INTERNAL VARIABLES */

const unsigned int bddRelOpCacheSize = BIG_TABLE; /* default = BIG */
const unsigned int bddRelOpSimpleCacheSize = BIG_TABLE; /* default = BIG */
const unsigned int bddRelOpComplexCacheSize = BIG_TABLE; /* default = BIG */
const unsigned int bddStabRSCacheSize = BIG_TABLE; /* default = BIG */

BddRelOpCacheTable bddRelOpCache
                   = {NULL,0,0,0,0,0};                     /* user cache */
BddRelOpSimpleCacheTable bddRelOpSimpleCache
                         = {NULL,0,0,0,0,0};               /* user cache */
BddRelOpComplexCacheTable bddRelOpComplexCache
                          = {NULL,0,0,0,0,0};              /* user cache */
BddStabRSCacheTable bddStabRSCache
                    = {NULL,0,0,0,0};                    /* user cache */

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static void RelOpGarbage(Biddy_Manager MNG);

static void RelOpSimpleGarbage(Biddy_Manager MNG);

static void RelOpComplexGarbage(Biddy_Manager MNG);

static void StabRSGarbage(Biddy_Manager MNG);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/* To simplify building on MS WINDOWS - Bdd_Init is stated in *.def file */
/* This funcion is defined in tcl part of the package, only */
/* Here, we make it always defined it even if we do not need it */

#ifndef TCL
   void Bdd_Init(void) {}
#endif

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_InitPkg()
{
    int i;

    if (bdd_status == 1) {
      printf("\nBDD package is already initialized. ");
      return;
    } else {
      bdd_status = 1;
    }

    Biddy_Init();

    /* INITIALIZATION OF TERMINAL EDGES */
    bdd_termTrue.p = (void *) Biddy_GetConstantOne();
    bdd_termTrue.mark = Biddy_GetMark(Biddy_GetConstantOne());
    bdd_termFalse.p = (void *) Biddy_GetConstantZero();
    bdd_termFalse.mark = Biddy_GetMark(Biddy_GetConstantZero());
    bdd_termNull.p = (void *) NULL;
    bdd_termNull.mark = FALSE;

    /* INITIALIZATION OF RelOp CACHE - USED FOR RELATIONAL OPERATIONS */
    bddRelOpCache.size = bddRelOpCacheSize;
    bddRelOpCache.num = 0;
    bddRelOpCache.search = 0;
    bddRelOpCache.find = 0;
    bddRelOpCache.overwrite = 0;
    if (!(bddRelOpCache.table = (BddRelOpCache *)
          malloc((bddRelOpCache.size+1) * sizeof(BddRelOpCache)))) {
      fprintf(stdout,"Bdd_InitPkg (RelOp): Out of memoy!\n");
      exit(1);
    }
    for (i=0;i<=bddRelOpCache.size;i++) {
      bddRelOpCache.table[i].key = 0;
    }
    Biddy_AddCache((Biddy_GCFunction) RelOpGarbage);

    /* INITIALIZATION OF RelOpSimple CACHE - SIMPLIFIED RELATIONAL OPERATIONS */
    bddRelOpSimpleCache.size = bddRelOpSimpleCacheSize;
    bddRelOpSimpleCache.num = 0;
    bddRelOpSimpleCache.search = 0;
    bddRelOpSimpleCache.find = 0;
    bddRelOpSimpleCache.overwrite = 0;
    if (!(bddRelOpSimpleCache.table = (BddRelOpSimpleCache *)
          malloc((bddRelOpSimpleCache.size+1) * sizeof(BddRelOpSimpleCache)))) {
      fprintf(stdout,"Bdd_InitPkg (RelOpSimple): Out of memoy!\n");
      exit(1);
    }
    for (i=0;i<=bddRelOpSimpleCache.size;i++) {
      bddRelOpSimpleCache.table[i].key = 0;
    }
    Biddy_AddCache((Biddy_GCFunction) RelOpSimpleGarbage);

    /* INITIALIZATION OF RelOpComplex CACHE - ADVANCED RELATIONAL OPERAT. */
    bddRelOpComplexCache.size = bddRelOpComplexCacheSize;
    bddRelOpComplexCache.num = 0;
    bddRelOpComplexCache.search = 0;
    bddRelOpComplexCache.find = 0;
    bddRelOpComplexCache.overwrite = 0;
    if (!(bddRelOpComplexCache.table = (BddRelOpComplexCache *)
          malloc((bddRelOpComplexCache.size+1) * sizeof(BddRelOpComplexCache)))) {
      fprintf(stdout,"Bdd_InitPkg (RelOpComplex): Out of memoy!\n");
      exit(1);
    }
    for (i=0;i<=bddRelOpComplexCache.size;i++) {
      bddRelOpComplexCache.table[i].key = 0;
    }
    Biddy_AddCache((Biddy_GCFunction) RelOpComplexGarbage);

    /* INITIALIZATION OF StabRS CACHE - USED FOR PARALLEL COMPOSITION */
    bddStabRSCache.size = bddStabRSCacheSize;
    bddStabRSCache.search = 0;
    bddStabRSCache.find = 0;
    bddStabRSCache.overwrite = 0;
    if (!(bddStabRSCache.table = (BddStabRSCache *)
          malloc((bddStabRSCache.size+1) * sizeof(BddStabRSCache)))) {
      fprintf(stdout,"Bdd_InitPkg (StabRS): Out of memoy!\n");
      exit(1);
    }
    for (i=0;i<=bddStabRSCache.size;i++) {
      bddStabRSCache.table[i].r = bdd_termNull;
    }
    Biddy_AddCache((Biddy_GCFunction) StabRSGarbage);

    /* DEBUGGING */
    /*
    fprintf(stderr,"TRUE: %u\n",TRUE);
    fprintf(stderr,"FALSE: %u\n",FALSE);
    */

}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_ExitPkg()
{
  int i;

  if (bdd_status == 0) {
    printf("\nBDD package is not initialized. ");
    return;
  } else {
    bdd_status = 0;
  }

  /* DEBUGGING */
  /*
  fprintf(stderr,"RELOP NUM: %llu\n",bddRelOpCache.num);
  fprintf(stderr,"RELOP SEARCH: %llu\n",bddRelOpCache.search);
  fprintf(stderr,"RELOP FIND: %llu\n",bddRelOpCache.find);
  fprintf(stderr,"RELOP OVERWRITE: %llu\n",bddRelOpCache.overwrite);
  */

  Biddy_Exit();

  /* DELETE RelOp CACHE TABLE */
  /*
  printf("Delete RelOp cache...\n");
  */
  free(bddRelOpCache.table);

  /* DELETE RelOpSimple CACHE TABLE */
  /*
  printf("Delete RelOpSimple cache...\n");
  */
  free(bddRelOpSimpleCache.table);

  /* DELETE RelOpComplex CACHE TABLE */
  /*
  printf("Delete RelOpComplex cache...\n");
  */
  for (i=0; i<bddRelOpComplexCache.size; i++) {
    if (bddRelOpComplexCache.table[i].key) {
      free(bddRelOpComplexCache.table[i].name);
    }
  }
  free(bddRelOpComplexCache.table);

}

/**Function****************************************************************
  Synopsis    []
  Description [Bdd_AboutPkg reports type and version of BDD package.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_AboutPkg()
{
  printf("EST ");
  printf(EDITION);
  printf(", version ");
  printf(VERSION);
  printf(", ");
  printf("Bdd package");
  printf(NEWLINE);
}

/**Function****************************************************************
  Synopsis    [Bdd_GetThen]
  Description []
  SideEffects [Input mark is not considered! To get correct subfuction
               use Bdd_TransferMark(Bdd_GetThen(f),f.mark).]
  SeeAlso     []
  Biddy       [Biddy_GetThen]
  CUDD        [Cudd_T]
  ************************************************************************/

Bdd_Edge Bdd_GetThen(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_GetThen((Biddy_Edge) f.p);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

  /* CUDD ?? NOT TESTED */
  /*
  Bdd_Edge r;
  r.p = (void *) Cudd_T((DdNode) f.p);
  r.mark = Cudd_IsComplement((DdNode) r.p);
  return r;
  */

}

/**Function****************************************************************
  Synopsis    [Bdd_GetElse]
  Description []
  SideEffects [Input mark is not considered! To get correct subfuction
               use Bdd_TransferMark(Bdd_GetElse(f),f.mark).]
  SeeAlso     []
  Biddy       [Biddy_GetElse]
  CUDD        [Cudd_E]
  ************************************************************************/

Bdd_Edge Bdd_GetElse(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_GetElse((Biddy_Edge) f.p);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

  /* CUDD ?? NOT TESTED */
  /*
  Bdd_Edge r;
  r.p = (void *) Cudd_E((DdNode) f.p);
  r.mark = Cudd_IsComplement((DdNode) r.p);
  return r;
  */

}

/**Function****************************************************************
  Synopsis    [Bdd_GetVariable]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_GetVariable]
  CUDD        [-]
  ************************************************************************/

Bdd_Variable Bdd_GetVariable(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  return (Bdd_Variable) Biddy_GetTopVariable((Biddy_Edge) f.p);
  /**/

  /* CUDD ?? NOT TESTED */
  /*
  return (Bdd_Variable) (Cudd_Regular((DdNode) r.p))->index;
  */

}

/**Function****************************************************************
  Synopsis    [Bdd_GetVariableName]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

Est_String Bdd_GetVariableName(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  return (Est_String) Biddy_GetTopVariableName((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_GetIthVariableName]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

Est_String Bdd_GetIthVariableName(Bdd_Variable v)
{

  /* BIDDY SINCE 1.2 */
  /**/
  return (Est_String) Biddy_GetVariableName((Biddy_Variable) v);
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_GetVariableChar]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

char Bdd_GetVariableChar(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  return Biddy_GetTopVariableChar((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_IsSmaller]
  Description []
  SideEffects [Always using biddyActiveOrdering.]
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

Est_Boolean Bdd_IsSmaller(Bdd_Edge f, Bdd_Edge g)
{

  /* BIDDY SINCE 1.2 */
  return Biddy_IsSmaller(
    Biddy_GetTopVariable((Biddy_Edge) f.p),
    Biddy_GetTopVariable((Biddy_Edge) g.p));
}

/**Function****************************************************************
  Synopsis    [Bdd_IsSmaller]
  Description []
  SideEffects [Always using biddyActiveOrdering.]
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

Est_Boolean Bdd_IsVariableSmaller(Bdd_Variable fv, Bdd_Variable gv)
{

  /* BIDDY SINCE 1.2 */
  return Biddy_IsSmaller(
    (Biddy_Variable) fv, (Biddy_Variable) gv);
}

/**Function****************************************************************
  Synopsis    [Bdd_Terminal]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_Terminal(Bdd_Variable v)
{

  /* BIDDY SINCE 1.2 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_FoaNode((Biddy_Variable) v, Biddy_GetConstantZero(), Biddy_GetConstantOne(), TRUE);
  r.mark = FALSE;
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_FoaTerminal]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_FoaTerminal]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_FoaTerminal(Est_String x)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_FoaVariable((Biddy_String) x);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_NOT]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NOT]
  CUDD        [Cudd_Not]
  ************************************************************************/

Bdd_Edge Bdd_NOT(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_NOT((Biddy_Edge) f.p);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_Eval1]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Eval1]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_Eval1(Est_String s)
{

  /* BIDDY SINCE 1.3 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_Eval1((Biddy_String) s);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_Eval2]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Eval2]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_Eval2(Est_String s)
{

  /* BIDDY SINCE 1.3 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_Eval2((Biddy_String) s);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_TransferMark]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_TransferMark]
  CUDD        [Cudd_NotCond]
  ************************************************************************/

Bdd_Edge Bdd_TransferMark(Bdd_Edge f, Est_Boolean mark)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_TransferMark((Biddy_Edge) f.p, (Biddy_Boolean) mark);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_ITE]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_ITE]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_ITE(Bdd_Edge f, Bdd_Edge g, Bdd_Edge h)
{
  Bdd_Edge r;

  /* DEBUGGING */
  /*
  if (!(f.p)) {
    fprintf(stderr,"ERROR (Bdd_ITE): f == NULL!\n");
    free((void *)1);
    exit(1);
  }
  if (!(g.p)) {
    fprintf(stderr,"ERROR (Bdd_ITE): g == NULL!\n");
    free((void *)1);
    exit(1);
  }
  if (!(h.p)) {
    fprintf(stderr,"ERROR (Bdd_ITE): h == NULL!\n");
    free((void *)1);
    exit(1);
  }
  if (!BddisOK(f)) {
    fprintf(stderr,"ERROR (Bdd_ITE): Bad f!\n");
    free((void *)1);
    exit(1);
  }
  if (!BddisOK(g)) {
    fprintf(stderr,"ERROR (Bdd_ITE): Bad g!\n");
    free((void *)1);
    exit(1);
  }
  if (!BddisOK(h)) {
    fprintf(stderr,"ERROR (Bdd_ITE): Bad h!\n");
    free((void *)1);
    exit(1);
  }
  */

  /* BIDDY SINCE 1.1 */
  /**/
  r.p = (void *) Biddy_ITE((Biddy_Edge) f.p, (Biddy_Edge) g.p, (Biddy_Edge) h.p);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  /**/

  return r;
}

/**Function****************************************************************
  Synopsis    [Bdd_Restrict]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Restrict]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_Restrict(Bdd_Edge f, Bdd_Variable v, Est_Boolean value)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_Restrict((Biddy_Edge) f.p, (Biddy_Variable) v, (value==TRUE) ? Biddy_GetConstantOne() : Biddy_GetConstantZero());
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_Compose]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Compose]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_Compose(Bdd_Edge f, Bdd_Variable v, Bdd_Edge g)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_Compose((Biddy_Edge) f.p, (Biddy_Variable) v, (Biddy_Edge) g.p);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_E]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_E]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_E(Bdd_Edge f, Bdd_Variable v)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_E((Biddy_Edge) f.p, (Biddy_Variable) v);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_A]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_A]
  CUDD        []
  ************************************************************************/

Bdd_Edge Bdd_A(Bdd_Edge f, Bdd_Variable v)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_A((Biddy_Edge) f.p, (Biddy_Variable) v);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_Clean]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Clean]
  CUDD        []
  ************************************************************************/

void Bdd_Clean()
{

  /* BIDDY SINCE 1.5 */
  /**/
  Biddy_Clean();
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_Garbage]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Garbage]
  CUDD        []
  ************************************************************************/

void Bdd_Garbage()
{

  /* BIDDY SINCE 1.6 */
  /**/
  Biddy_FullGC();
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_Refresh]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Refresh]
  CUDD        []
  ************************************************************************/

void Bdd_Refresh(Bdd_Edge f)
{

  /* BIDDY SINCE 1.5 */
  /**/
  Biddy_Refresh((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [Bdd_Reorder]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Reorder]
  CUDD        []
  ************************************************************************/

void Bdd_Reorder(Bdd_Edge f)
{

  /* BIDDY SINCE 1.5 */
  /**/
  if (Bdd_isNull(f)) {
    Biddy_PurgeAndReorder(NULL,FALSE);
  } else {
    Biddy_PurgeAndReorder((Biddy_Edge) f.p, FALSE);
  }
  /**/

}

/**Function****************************************************************
  Synopsis    [Function Bdd_FindFormula]
  Description [Bdd_FindFormula find formula in Biddy Formula table.]
  SideEffects []
  SeeAlso     [Bdd_AddFormula]
  ************************************************************************/

Est_Boolean
Bdd_FindFormula(Est_String x, Bdd_Edge *f)
{
  Est_Boolean ok;

  ok = Biddy_FindFormula((Biddy_String) x, (Biddy_Edge *) &(f->p));
  if (ok) {
    f->mark = Biddy_GetMark((Biddy_Edge) f->p);
  }

  return ok;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_SaveFormula]
  Description [Bdd_SaveFormula adds formula to the Biddy Formula table.]
  SideEffects [If prefix!=NULL it will be added to the name.
               If name==NULL then pointer f is converted to string and used.]
  SeeAlso     []
  ************************************************************************/

void
Bdd_SaveFormula(Est_String prefix, Est_String name, Bdd_Edge f)
{
  Est_String s;

  if (prefix) {
    if (name) {
      s = (Est_String) malloc(strlen(prefix) + strlen(name) + 1);
      sprintf(s,"%s%s",prefix,name);
    } else {
      s = (Est_String) malloc(strlen(prefix) + 18 + 1); /* 64-bit address */
      sprintf(s,"%s%p",prefix,(void *)f.p);
    }
  } else {
    if (name) {
      s = strdup(name);
    } else {
      s = (Est_String) malloc(18 + 1); /* 64-bit address */
      sprintf(s,"%p",(void *)f.p);
    }
  }

  Biddy_AddPersistentFormula((Biddy_String) s, (Biddy_Edge) f.p);

  free(s);
}

/**Function****************************************************************
  Synopsis    [Function Bdd_SaveTmpFormula]
  Description [Bdd_SaveTmpFormula adds formula to the Biddy Formula table.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_SaveTmpFormula(Bdd_Edge f, int count)
{
  Biddy_AddTmpFormula((Biddy_Edge) f.p, count);
}

/**Function****************************************************************
  Synopsis    [Function Bdd_WriteBDD.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_WriteBDD(Est_String name)
{
  Bdd_Edge f;

  if (!Bdd_FindFormula(name,&f)) {
    printf("Function %s does not exists!\n",name);
    return;
  } else {
    Biddy_WriteBDD((Biddy_Edge) f.p);
  }
}

/**Function****************************************************************
  Synopsis    [Function Bdd_WriteTable.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_WriteTable(Est_String name)
{
  Bdd_Edge f;

  if (!Bdd_FindFormula(name,&f)) {
    printf("Function %s does not exists!\n",name);
    return;
  } else {
    Biddy_WriteTable((Biddy_Edge) f.p);
  }
}

/**Function****************************************************************
  Synopsis    [Function Bdd_WriteDot.]
  Description []
  SideEffects [if (id != -1) use it instead of <...>]
  SeeAlso     []
  ************************************************************************/

int
Bdd_WriteDot(char *filename, Est_String name, int id)
{
  Bdd_Edge f;

  if (!Bdd_FindFormula(name,&f)) {
    printf("Function %s does not exists!\n",name);
    return(0);
  }

  return Biddy_WriteDotx(filename,(Biddy_Edge) f.p, name,id);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_SystemStat(FILE *s)
{
  if (!s || (s == stdout)) {
    Biddy_PrintInfo(NULL);
  } else {
    Biddy_PrintInfo(s);
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_FunctionStat(FILE *s, Bdd_Edge f)
{
  int num;

  num = BddNodeNumber(f);

  if (!s || (s == stdout)) {
    printf("%d/",num);
    printf("%d/",BddNodeMaxLevel(f));
    printf("%f\n",BddNodeAvgLevel(f));
  } else {
    fprintf(s,"%d/",num);
    fprintf(s,"%d/",BddNodeMaxLevel(f));
    fprintf(s,"%f\n",BddNodeAvgLevel(f));
  }
}

/**Function****************************************************************
  Synopsis    [Bdd_Support]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_Support]
  CUDD        [Cudd_Support]
  ************************************************************************/

Bdd_Edge Bdd_Support(Bdd_Edge f)
{

  /* BIDDY SINCE 1.4 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_Support((Biddy_Edge) f.p);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [BddisOK]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_IsOK]
  CUDD        []
  ************************************************************************/

Est_Boolean BddisOK(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  return (Est_Boolean) Biddy_IsOK((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddSelectNode]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_SelectNode]
  CUDD        []
  ************************************************************************/

void BddSelectNode(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Biddy_SelectNode((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddDeselectNode]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_DeselectNode]
  CUDD        []
  ************************************************************************/

void BddDeselectNode(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  Biddy_DeselectNode((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddisSelected]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_IsSelected]
  CUDD        []
  ************************************************************************/

Est_Boolean BddisSelected(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  return (Est_Boolean) Biddy_IsSelected((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeSelect]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeSelect]
  CUDD        []
  ************************************************************************/

void BddNodeSelect(Bdd_Edge f)
{

  /* BIDDY SINCE 1.6 */
  /**/
  Biddy_SelectNode(Biddy_GetConstantOne());
  Biddy_SelectFunction((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeRepair]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeRepair]
  CUDD        []
  ************************************************************************/

void BddNodeRepair(Bdd_Edge f)
{

  /* BIDDY SINCE 1.6 */
  /**/
  Biddy_DeselectAll();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddFoaNode]
  Description []
  SideEffects [Using BddFoaNode you can create node with bad ordering.
               It is much more safe to use Bdd_ITE!]
  SeeAlso     []
  Biddy       [Biddy_FoaNode]
  CUDD        [cuddUniqueInter]
  ************************************************************************/

Bdd_Edge BddFoaNode(Bdd_Variable v, Bdd_Edge pf, Bdd_Edge pt)
{

  /* BIDDY SINCE 1.2 */
  /**/
  Bdd_Edge r;
  r.p = (void *) Biddy_FoaNode((Biddy_Variable) v, (Biddy_Edge) pf.p, (Biddy_Edge) pt.p, TRUE);
  r.mark = Biddy_GetMark((Biddy_Edge) r.p);
  return r;
  /**/

}

/**Function****************************************************************
  Synopsis    [BddExchange]
  Description [BddExchange exchange two Boolean functions.]
  SideEffects []
  SeeAlso     []
  Biddy       []
  CUDD        []
  ************************************************************************/

void
BddExchange(Bdd_Edge *f, Bdd_Edge *g)
{
  Bdd_Edge sup;

  /* BIDDY SINCE 1.0 */
  sup = *f;
  *f = *g;
  *g = sup;
}

/**Function****************************************************************
  Synopsis    [BddNodeTableSize]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableSize]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeTableSize()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_NodeTableSize();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeTableMax]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableMax]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeTableMax()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_NodeTableMax();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeTableNum]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableNum]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeTableNum()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_NodeTableNum();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeTableGarbage]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableGarbage]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeTableGarbage()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_NodeTableGCNumber();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeTableGenerated]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableGenerated]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeTableGenerated()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_NodeTableGenerated();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeTableBlockNumber]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableBlockNumber]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeTableBlockNumber()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_NodeTableBlockNumber();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeTableSwapNumber]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableSwapNumber]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeTableSwapNumber()
{

  /* BIDDY SINCE 1.2 */
  /**/
  return Biddy_NodeTableSwapNumber();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeTableSiftingNumber]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeTableSiftingNumber]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeTableSiftingNumber()
{

  /* BIDDY SINCE 1.2 */
  /**/
  return Biddy_NodeTableSiftingNumber();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddListUsed]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_ListUsed]
  CUDD        []
  ************************************************************************/

unsigned int BddListUsed()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_ListUsed();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddListMaxLength]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_ListMaxLength]
  CUDD        []
  ************************************************************************/

unsigned int BddListMaxLength()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_ListMaxLength();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddListAvgLength]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_ListAvgLength]
  CUDD        []
  ************************************************************************/

float BddListAvgLength()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_ListAvgLength();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddIteCacheSearch]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_IteCacheSearch]
  CUDD        []
  ************************************************************************/

unsigned long long int BddIteCacheSearch()
{

  /* BIDDY SINCE 1.2 */
  /**/
  return Biddy_IteCacheSearch();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddIteCacheFind]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_IteCacheFind]
  CUDD        []
  ************************************************************************/

unsigned long long int BddIteCacheFind()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_IteCacheFind();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddIteCacheOverwrite]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_IteCacheOverwrite]
  CUDD        []
  ************************************************************************/

unsigned long long int BddIteCacheOverwrite()
{

  /* BIDDY SINCE 1.0 */
  /**/
  return Biddy_IteCacheOverwrite();
  /**/

}

/**Function****************************************************************
  Synopsis    [BddRelopCacheNum]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopCacheNum()
{
  return bddRelOpCache.num;
}

/**Function****************************************************************
  Synopsis    [BddRelopCacheSearch]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopCacheSearch()
{
  return bddRelOpCache.search;
}

/**Function****************************************************************
  Synopsis    [BddRelopCacheFind]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopCacheFind()
{
  return bddRelOpCache.find;
}

/**Function****************************************************************
  Synopsis    [BddRelopCacheOverwrite]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopCacheOverwrite()
{
  return bddRelOpCache.overwrite;
}

/**Function****************************************************************
  Synopsis    [BddRelopSimpleCacheNum]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopSimpleCacheNum()
{
  return bddRelOpSimpleCache.num;
}

/**Function****************************************************************
  Synopsis    [BddRelopSimpleCacheSearch]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopSimpleCacheSearch()
{
  return bddRelOpSimpleCache.search;
}

/**Function****************************************************************
  Synopsis    [BddRelopSimpleCacheFind]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopSimpleCacheFind()
{
  return bddRelOpSimpleCache.find;
}

/**Function****************************************************************
  Synopsis    [BddRelopSimpleCacheOverwrite]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopSimpleCacheOverwrite()
{
  return bddRelOpSimpleCache.overwrite;
}

/**Function****************************************************************
  Synopsis    [BddRelopComplexCacheNum]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopComplexCacheNum()
{
  return bddRelOpComplexCache.num;
}

/**Function****************************************************************
  Synopsis    [BddRelopComplexCacheSearch]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopComplexCacheSearch()
{
  return bddRelOpComplexCache.search;
}

/**Function****************************************************************
  Synopsis    [BddRelopComplexCacheFind]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopComplexCacheFind()
{
  return bddRelOpComplexCache.find;
}

/**Function****************************************************************
  Synopsis    [BddRelopComplexCacheOverwrite]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddRelopComplexCacheOverwrite()
{
  return bddRelOpComplexCache.overwrite;
}

/**Function****************************************************************
  Synopsis    [BddStabRSCacheSearch]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddStabRSCacheSearch()
{
  return bddStabRSCache.search;
}

/**Function****************************************************************
  Synopsis    [BddStabRSCacheFind]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddStabRSCacheFind()
{
  return bddStabRSCache.find;
}

/**Function****************************************************************
  Synopsis    [BddStabRSCacheOverwrite]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

unsigned long long int BddStabRSCacheOverwrite()
{
  return bddStabRSCache.overwrite;
}

/**Function****************************************************************
  Synopsis    [BddNodeNumber]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeNumber]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeNumber(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  return Biddy_NodeNumber((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeNumberPlain]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeNumberPlain]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeNumberPlain(Bdd_Edge f)
{

  /* BIDDY SINCE 1.4 */
  /**/
  return Biddy_NodeNumberPlain((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeMaxLevel]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeMaxLevel]
  CUDD        []
  ************************************************************************/

unsigned int BddNodeMaxLevel(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  return Biddy_NodeMaxLevel((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddNodeAvgLevel]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_NodeAvgLevel]
  CUDD        []
  ************************************************************************/

float BddNodeAvgLevel(Bdd_Edge f)
{

  /* BIDDY SINCE 1.1 */
  /**/
  return Biddy_NodeAvgLevel((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddVariableNumber]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_VariableNumber]
  CUDD        []
  ************************************************************************/

unsigned int BddVariableNumber(Bdd_Edge f)
{

  /* BIDDY SINCE 1.4 */
  /**/
  return Biddy_VariableNumber((Biddy_Edge) f.p);
  /**/

}

/**Function****************************************************************
  Synopsis    [BddMintermCount]
  Description []
  SideEffects []
  SeeAlso     []
  Biddy       [Biddy_MintermCount]
  CUDD        []
  ************************************************************************/

double
BddMintermCount(Bdd_Edge f, int nvars)
{

  /* BIDDY SINCE 1.4 */
  /**/
  return Biddy_CountMinterm((Biddy_Edge) f.p, nvars);
  /**/

}

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function RelOpGarbage.]
  Description []
  SideEffects [Garbage collection always depends on a manager.
               It is not possible to use unmanaged functions here!]
  SeeAlso     []
  ************************************************************************/

static void
RelOpGarbage(Biddy_Manager MNG)
{
  int i;
  BddRelOpCache *p;

  for (i=0; i<=bddRelOpCache.size; i++) {
    if (bddRelOpCache.table[i].key) {
      p = &bddRelOpCache.table[i];
      if (!Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->f).p)) ||
          !Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->g).p)) ||
          !Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->r).p)))
      {
        p->key = 0;
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    [Function RelOpSimpleGarbage.]
  Description []
  SideEffects [Garbage collection always depends on a manager.
               It is not possible to use unmanaged functions here!]
  SeeAlso     []
  ************************************************************************/

static void
RelOpSimpleGarbage(Biddy_Manager MNG)
{
  int i;
  BddRelOpSimpleCache *p;

  for (i=0; i<=bddRelOpSimpleCache.size; i++) {
    if (bddRelOpSimpleCache.table[i].key) {
      p = &bddRelOpSimpleCache.table[i];
      if (!Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->f).p)) ||
          !Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->r).p)))
      {
        p->key = 0;
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    [Function RelOpComplexGarbage.]
  Description []
  SideEffects [Garbage collection always depends on a manager.
               It is not possible to use unmanaged functions here!]
  SeeAlso     []
  ************************************************************************/

static void
RelOpComplexGarbage(Biddy_Manager MNG)
{
  int i;
  BddRelOpComplexCache *p;

  for (i=0; i<=bddRelOpComplexCache.size; i++) {
    if (bddRelOpComplexCache.table[i].key) {
      p = &bddRelOpComplexCache.table[i];
      if (!Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->f).p)) ||
          !Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->r).p)))
      {
        free(p->name);
        p->key = 0;
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    [Function RelOpGarbage.]
  Description []
  SideEffects [Garbage collection always depends on a manager.
               It is not possible to use unmanaged functions here!]
  SeeAlso     []
  ************************************************************************/

static void
StabRSGarbage(Biddy_Manager MNG)
{
  int i;
  BddStabRSCache *p;

  for (i=0; i<=bddStabRSCache.size; i++) {
    if (!Bdd_isNull(bddStabRSCache.table[i].r)) {
      p = &bddStabRSCache.table[i];
      if (!Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->f).p)) ||
          !Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->stablist).p)) ||
          !Biddy_Managed_IsOK(MNG,(Biddy_Edge)((p->r).p)))
      {
        p->r = bdd_termNull;
      }
    }
  }
}
