/**CFile*******************************************************************
  PackageName [bdd]
  Synopsis    [Package 'bdd' enable symbolic computations by representing
               Boolean functions with ROBDDs.]

  FileName    [bddRel.c]
  Revision    [$Revision: 113 $]
  Date        [$Date: 2016-03-08 22:31:16 +0100 (tor, 08 mar 2016) $]
  Authors     [Robert Meolic (robert.meolic@um.si),
               Ales Casar (ales@homemade.net)]
  Description [The file bddRel.c contains functions for relational
               operations (e.g. RelProd).]
  SeeAlso     [bdd.h, bddInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 2003, 2016
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

/* Because of compatibility with old GLIBC we use memmove instead of memcpy. */
/* However, this approach is slower (maybe 10%, maybe more)! */

#include "bddInt.h"

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

#define oNULL 0
#define oAND 1
#define oOR 2
#define oIMPL 3

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [NOT REENTRANT! Ex and Ax must not be both used!]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_RelOp(Bdd_Edge f, Bdd_Edge g, Est_String format, Est_Boolean firsttime)
{

  BddRelOpCache *p; /* MUST NOT BE STATIC */
  Bdd_Edge f0, f1, g0, g1, h0, h1; /* MUST NOT BE STATIC */
  Bdd_Variable zvar; /* MUST NOT BE STATIC */
  char zchar; /* MUST NOT BE STATIC */

  Est_Boolean done; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */
  Bdd_Edge h; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */

  static int op;            /* MUST BE STATIC */

  static Est_Boolean ExA;   /* MUST BE STATIC */
  static Est_Boolean AxA;   /* MUST BE STATIC */
  static Est_Boolean ExB;   /* MUST BE STATIC */
  static Est_Boolean AxB;   /* MUST BE STATIC */
  static Est_Boolean ExR;   /* MUST BE STATIC */
  static Est_Boolean AxR;   /* MUST BE STATIC */
  static Est_Boolean ExS;   /* MUST BE STATIC */
  static Est_Boolean AxS;   /* MUST BE STATIC */
  static Est_Boolean ExP;   /* MUST BE STATIC */
  static Est_Boolean AxP;   /* MUST BE STATIC */
  static Est_Boolean ExQ;   /* MUST BE STATIC */
  static Est_Boolean AxQ;   /* MUST BE STATIC */

  static Est_Boolean A2B;   /* MUST BE STATIC */
  static Est_Boolean R2S;   /* MUST BE STATIC */
  static Est_Boolean R2P;   /* MUST BE STATIC */
  static Est_Boolean R2Q;   /* MUST BE STATIC */
  static Est_Boolean S2R;   /* MUST BE STATIC */
  static Est_Boolean S2P;   /* MUST BE STATIC */
  static Est_Boolean S2Q;   /* MUST BE STATIC */
  static Est_Boolean P2Q;   /* MUST BE STATIC */
  static Est_Boolean P2R;   /* MUST BE STATIC */
  static Est_Boolean Q2P;   /* MUST BE STATIC */

  static unsigned int key;  /* CACHE TABLE - MUST BE STATIC */

  /* IF FIRST-TIME, PARSING FORMAT STRING */

  if (firsttime) {

    /* DEBUGGING */
    /*
    if (!BddisOK(f)) {
      fprintf(stderr,"ERROR (Bdd_RelOp): Bad f!\n");
      free((void *)1);
      exit(1);
    }
    if (!BddisOK(g)) {
      fprintf(stderr,"ERROR (Bdd_RelOp): Bad g!\n");
      free((void *)1);
      exit(1);
    }
    */

    bddRelOpCache.num++;

    key = 0;                               /* 0 * 2^30 */
    op = oNULL;

    if (strstr(format,"#AND")) {
      key = 1073741824U;                   /* 1 * 2^30 */
      op = oAND;
    }

    if (strstr(format,"#OR")) {
      key = 2147483648U;                   /* 2 * 2^30 */
      op = oOR;
    }

    if (strstr(format,"#IMPL")) {
      key = 3221225472U;                   /* 3 * 2^30 */
      op = oIMPL;
    }

    if (strstr(format,"Ex") != NULL) {
      ExA = (strstr(format,"xA") != NULL);
      ExB = (strstr(format,"xB") != NULL);
      ExR = (strstr(format,"xR") != NULL);
      ExS = (strstr(format,"xS") != NULL);
      ExP = (strstr(format,"xP") != NULL);
      ExQ = (strstr(format,"xQ") != NULL);
    } else {
      ExA = FALSE;
      ExB = FALSE;
      ExR = FALSE;
      ExS = FALSE;
      ExP = FALSE;
      ExQ = FALSE;
    }

    if (strstr(format,"Ax") != NULL) {
      AxA = (strstr(format,"xA") != NULL);
      AxB = (strstr(format,"xB") != NULL);
      AxR = (strstr(format,"xR") != NULL);
      AxS = (strstr(format,"xS") != NULL);
      AxP = (strstr(format,"xP") != NULL);
      AxQ = (strstr(format,"xQ") != NULL);
    } else {
      AxA = FALSE;
      AxB = FALSE;
      AxR = FALSE;
      AxS = FALSE;
      AxP = FALSE;
      AxQ = FALSE;
    }

    A2B = (strstr(format,"A2B") != NULL);
    R2S = (strstr(format,"R2S") != NULL);
    R2P = (strstr(format,"R2P") != NULL);
    R2Q = (strstr(format,"R2Q") != NULL);
    S2R = (strstr(format,"S2R") != NULL);
    S2P = (strstr(format,"S2P") != NULL);
    S2Q = (strstr(format,"S2Q") != NULL);
    P2Q = (strstr(format,"P2Q") != NULL);
    P2R = (strstr(format,"P2R") != NULL);
    Q2P = (strstr(format,"Q2P") != NULL);

    if (ExA) key = key + 1;
    if (AxA) key = key + 2;
    if (ExB) key = key + 4;
    if (AxB) key = key + 8;
    if (ExR) key = key + 16;
    if (AxR) key = key + 32;
    if (ExS) key = key + 64;
    if (AxS) key = key + 128;
    if (ExP) key = key + 256;
    if (AxP) key = key + 512;
    if (ExQ) key = key + 1024;
    if (AxQ) key = key + 2048;

    if (A2B) key = key + 4096;
    if (R2S) key = key + 8192;
    if (R2P) key = key + 16384;
    if (R2Q) key = key + 32768;
    if (S2R) key = key + 65536;
    if (S2P) key = key + 131072;
    if (S2Q) key = key + 262144;
    if (P2Q) key = key + 524288;
    if (P2R) key = key + 1048576;
    if (Q2P) key = key + 2097152;

    h = Bdd_RelOp(f,g,format,FALSE);

  }

  if (!firsttime) {

    if (op == oNULL) {
      if (Bdd_isTerminal(f))
         return f;
    }

    if (op == oAND) {
      if (Bdd_isEqv(f, bdd_termFalse) || Bdd_isEqv(g, bdd_termFalse))
        return bdd_termFalse;
      if (Bdd_isEqv(f, bdd_termTrue) && Bdd_isEqv(g, bdd_termTrue))
        return bdd_termTrue;
    }

    if (op == oOR) {
      if (Bdd_isEqv(f, bdd_termTrue) || Bdd_isEqv(g, bdd_termTrue))
        return bdd_termTrue;
      if (Bdd_isEqv(f, bdd_termFalse) && Bdd_isEqv(g, bdd_termFalse))
        return bdd_termFalse;
    }

    if (op == oIMPL) {
      if (Bdd_isEqv(f, bdd_termFalse) || Bdd_isEqv(g, bdd_termTrue))
        return bdd_termTrue;
      if (Bdd_isEqv(f, bdd_termTrue) && Bdd_isEqv(g, bdd_termFalse))
        return bdd_termFalse;
    }

    /* IF COMUTATIVE OPERATION, DO SIMPLE NORMALIZATION */
    if ((op == oOR) || (op == oAND)) {
      if ((uintptr_t) f.p > (uintptr_t) g.p) BddExchange(&f, &g);
    }

    /* IS ANOTHER NORMALIZATION POSSIBLE ??? */

    /* LOOKUP IN CACHE TABLE */

    bddRelOpCache.search++;

    /* NAIVE HASH FUNCTION */
    /*
    p = &bddRelOpCache.table[(
          ((uintptr_t) key << 0) +
          ((uintptr_t) f.p << 0) +
          ((uintptr_t) g.p << 3)
		    ) & bddRelOpCache.size];
    */

#if UINTPTR_MAX == 0xffffffffffffffff

    /* MURMUR3 (A KIND OF) 64-bit HASH FUNCTION */
    /* https://code.google.com/p/smhasher/ */
    {
    const uint64_t C1 = 0x87c37b91114253d5L;
    const uint64_t C2 = 0x4cf5ad432745937fL;
    uintptr_t k1,k2,k;
    k = key;
    k1 = (uintptr_t) f.p;
    k1 *= C1;
    k1 = (k1 << 31) | (k1 >> 33);
    k1 *= C2;
    k ^= k1;
    k = (k << 27) | (k >> 37);
    k = k*5+0x52dce729;
    k2 = (uintptr_t) g.p;
    k2 *= C2;
    k2 = (k2 << 33) | (k2 >> 31);
    k2 *= C1;
    k ^= k2;
    k = (k << 31) | (k >> 33);
    k = k*5+0x38495ab5;
    k ^= k >> 33;
    k *= 0xff51afd7ed558ccdL;
    k ^= k >> 33;
    k *= 0xc4ceb9fe1a85ec53L;
    k ^= k >> 33;
    p = &bddRelOpCache.table[k & bddRelOpCache.size];
    }

#else

    /* MURMUR3 (A KIND OF) 32-bit HASH FUNCTION */
    /* https://code.google.com/p/smhasher/ */
    {
    const uint32_t C1 = 0xcc9e2d51;
    const uint32_t C2 = 0x1b873593;
    uintptr_t k1,k2,k;
    k = key;
    k1 = (uintptr_t) f.p;
    k1 *= C1;
    k1 = (k1 << 15) | (k1 >> 17);
    k1 *= C2;
    k ^= k1;
    k = (k << 13) | (k >> 19);
    k = k*5+0xe6546b64;
    k2 = (uintptr_t) g.p;
    k2 *= C1;
    k2 = (k2 << 15) | (k2 >> 17);
    k2 *= C2;
    k ^= k2;
    k = (k << 13) | (k >> 19);
    k = k*5+0xe6546b64;
    k ^= k >> 16;
    k *= 0x85ebca6b;
    k ^= k >> 13;
    k *= 0xc2b2ae35;
    k ^= k >> 16;
    p = &bddRelOpCache.table[k & bddRelOpCache.size];
    }

#endif

    if ((p->key == key) &&
        Bdd_isEqv(p->f,f) &&
        Bdd_isEqv(p->g,g))
    {
      bddRelOpCache.find++;
      h = p->r;
      Bdd_Refresh(h);
      return h;
    }

    /* DETERMINING THE SUCCESSORS AND TOP VARIABLE */

    if (op == oNULL) {
      f0 = Bdd_TransferMark(Bdd_GetElse(f), f.mark);
      f1 = Bdd_TransferMark(Bdd_GetThen(f), f.mark);
      zvar = Bdd_GetVariable(f);
      zchar = Bdd_GetVariableChar(f);
    } else {
      if (Bdd_IsSmaller(f,g)) {
        f0 = Bdd_TransferMark(Bdd_GetElse(f), f.mark);
        f1 = Bdd_TransferMark(Bdd_GetThen(f), f.mark);
        g0 = g;
        g1 = g;
        zvar = Bdd_GetVariable(f);
        zchar = Bdd_GetVariableChar(f);
      } else if (Bdd_IsSmaller(g,f)) {
        f0 = f;
        f1 = f;
        g0 = Bdd_TransferMark(Bdd_GetElse(g), g.mark);
        g1 = Bdd_TransferMark(Bdd_GetThen(g), g.mark);
        zvar = Bdd_GetVariable(g);
        zchar = Bdd_GetVariableChar(g);
      } else {
        f0 = Bdd_TransferMark(Bdd_GetElse(f), f.mark);
        f1 = Bdd_TransferMark(Bdd_GetThen(f), f.mark);
        g0 = Bdd_TransferMark(Bdd_GetElse(g), g.mark);
        g1 = Bdd_TransferMark(Bdd_GetThen(g), g.mark);
        zvar = Bdd_GetVariable(f);
        zchar = Bdd_GetVariableChar(f);
      }
    }

    /* DO RECURSIVE CALLS */
    /* IN BOOLEAN OPERATORS ARE INVOLVED WE HAVE SOME OPTIMIZATIONS */
    /* ONLY ONE RECURSIVE CALL IS DONE HERE, THE OTHER ONE MAY NOT BE NEEDED! */

    if (op == oNULL) {
      h0 = Bdd_RelOp(f0, g, format, firsttime);
      h1 = Bdd_RelOp(f1, g, format, firsttime);
    } else {
      h0 = Bdd_RelOp(f0, g0, format, firsttime);
    }

    /* COMPOSE PARTIAL RESULTS */
    switch (zchar) {

    case 'a':
    {
      done = FALSE;

      if (ExA) {
        if (Bdd_isEqv(h0,bdd_termTrue)) {
          h = bdd_termTrue;
        } else if (Bdd_isEqv(h0,f1) || Bdd_isEqv(h0,g1)) {
          h = h0;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, bdd_termTrue, h1);
        }
        done = TRUE;
      }

      if (AxA) {
        if (Bdd_isEqv(h0,bdd_termFalse)) {
          h = bdd_termFalse;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, h1, bdd_termFalse);
        }
        done = TRUE;
      }

      if (A2B) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "ab", thus b=a+1 */
        done = TRUE;
      }

      if (!done) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'b':
    {
      done = FALSE;

      if (ExB) {
        if (Bdd_isEqv(h0,bdd_termTrue)) {
          h = bdd_termTrue;
        } else if (Bdd_isEqv(h0,f1) || Bdd_isEqv(h0,g1)) {
          h = h0;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, bdd_termTrue, h1);
        }
        done = TRUE;
      }

      if (AxB) {
        if (Bdd_isEqv(h0,bdd_termFalse)) {
          h = bdd_termFalse;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, h1, bdd_termFalse);
        }
        done = TRUE;
      }

      if (!done) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'r':
    {
      done = FALSE;

      if (ExR) {
        if (Bdd_isEqv(h0,bdd_termTrue)) {
          h = bdd_termTrue;
        } else if (Bdd_isEqv(h0,f1) || Bdd_isEqv(h0,g1)) {
          h = h0;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, bdd_termTrue, h1);
        }
        done = TRUE;
      }

      if (AxR) {
        if (Bdd_isEqv(h0,bdd_termFalse)) {
          h = bdd_termFalse;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, h1, bdd_termFalse);
        }
        done = TRUE;
      }

      if (R2S) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus s=r+1 */
        done = TRUE;
      }

      if (R2P) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar+2), h1, h0); /* "rspq", thus p=r+2 */
        done = TRUE;
      }

      if (R2Q) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar+3), h1, h0); /* "rspq", thus q=r+3 */
        done = TRUE;
      }

      if (!done) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 's':
    {
      done = FALSE;

      if (ExS) {
        if (Bdd_isEqv(h0,bdd_termTrue)) {
          h = bdd_termTrue;
        } else if (Bdd_isEqv(h0,f1) || Bdd_isEqv(h0,g1)) {
          h = h0;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, bdd_termTrue, h1);
        }
        done = TRUE;
      }

      if (AxS) {
        if (Bdd_isEqv(h0,bdd_termFalse)) {
          h = bdd_termFalse;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, h1, bdd_termFalse);
        }
        done = TRUE;
      }

      if (S2R) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar-1), h1, h0); /* "rspq", thus r=s-1 */
        done = TRUE;
      }

      if (S2P) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus p=s+1 */
        done = TRUE;
      }

      if (S2Q) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar+2), h1, h0); /* "rspq", thus q=s+2 */
        done = TRUE;
      }

      if (!done) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'p':
    {
      done = FALSE;

      if (ExP) {
        if (Bdd_isEqv(h0,bdd_termTrue)) {
          h = bdd_termTrue;
        } else if (Bdd_isEqv(h0,f1) || Bdd_isEqv(h0,g1)) {
          h = h0;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, bdd_termTrue, h1);
        }
        done = TRUE;
      }

      if (AxP) {
        if (Bdd_isEqv(h0,bdd_termFalse)) {
          h = bdd_termFalse;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, h1, bdd_termFalse);
        }
        done = TRUE;
      }

      if (P2Q) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus q=p+1 */
        done = TRUE;
      }

      if (P2R) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar-2), h1, h0); /* "rspq", thus r=p-2 */
        done = TRUE;
      }

      if (!done) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'q':
    {
      done = FALSE;

      if (ExQ) {
        if (Bdd_isEqv(h0,bdd_termTrue)) {
          h = bdd_termTrue;
        } else if (Bdd_isEqv(h0,f1) || Bdd_isEqv(h0,g1)) {
          h = h0;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, bdd_termTrue, h1);
        }
        done = TRUE;
      }

      if (AxQ) {
        if (Bdd_isEqv(h0,bdd_termFalse)) {
          h = bdd_termFalse;
        } else {
          h1 = Bdd_RelOp(f1, g1, format, firsttime);
          h = Bdd_ITE(h0, h1, bdd_termFalse);
        }
        done = TRUE;
      }

      if (Q2P) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar-1), h1, h0); /* "rspq", thus p=q-1 */
        done = TRUE;
      }

      if (!done) {
        h1 = Bdd_RelOp(f1, g1, format, firsttime);
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    } /* switch zchar */

    /* STORE IN CACHE TABLE */

    if (p->key) bddRelOpCache.overwrite++;
    p->f = f;
    p->g = g;
    p->r = h;
    p->key = key;

  }

  return h;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [NOT REENTRANT!]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_RelOpSimple(Bdd_Edge f, Est_String format, Est_Boolean firsttime)
{
  BddRelOpSimpleCache *p; /* MUST NOT BE STATIC */
  Bdd_Edge f0, f1, h0, h1; /* MUST NOT BE STATIC */
  Bdd_Variable zvar; /* MUST NOT BE STATIC */
  char zchar; /* MUST NOT BE STATIC */

  Est_Boolean done; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */
  Bdd_Edge h; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */

  static Est_Boolean ExA;   /* MUST BE STATIC */
  static Est_Boolean AxA;   /* MUST BE STATIC */
  static Est_Boolean ExB;   /* MUST BE STATIC */
  static Est_Boolean AxB;   /* MUST BE STATIC */
  static Est_Boolean ExR;   /* MUST BE STATIC */
  static Est_Boolean AxR;   /* MUST BE STATIC */
  static Est_Boolean ExS;   /* MUST BE STATIC */
  static Est_Boolean AxS;   /* MUST BE STATIC */
  static Est_Boolean ExP;   /* MUST BE STATIC */
  static Est_Boolean AxP;   /* MUST BE STATIC */
  static Est_Boolean ExQ;   /* MUST BE STATIC */
  static Est_Boolean AxQ;   /* MUST BE STATIC */

  static Est_Boolean A2B;   /* MUST BE STATIC */
  static Est_Boolean R2S;   /* MUST BE STATIC */
  static Est_Boolean R2P;   /* MUST BE STATIC */
  static Est_Boolean R2Q;   /* MUST BE STATIC */
  static Est_Boolean S2R;   /* MUST BE STATIC */
  static Est_Boolean S2P;   /* MUST BE STATIC */
  static Est_Boolean S2Q;   /* MUST BE STATIC */
  static Est_Boolean P2Q;   /* MUST BE STATIC */
  static Est_Boolean P2R;   /* MUST BE STATIC */
  static Est_Boolean Q2P;   /* MUST BE STATIC */

  static unsigned int key;  /* CACHE TABLE - MUST BE STATIC */

  if (Bdd_isTerminal(f)) return f;

  /* IF FIRST-TIME, PARSING FORMAT STRING */

  if (firsttime) {

    /* DEBUGGING */
    /*
    if (!BddisOK(f)) {
      fprintf(stderr,"ERROR (Bdd_RelOpSimple): Bad f!\n");
      free((void *)1);
      exit(1);
    }
    */

    bddRelOpSimpleCache.num++;

    key = 0;                               /* 0 * 2^30 */

    if (strstr(format,"Ex") != NULL) {
      ExA = (strstr(format,"xA") != NULL);
      ExB = (strstr(format,"xB") != NULL);
      ExR = (strstr(format,"xR") != NULL);
      ExS = (strstr(format,"xS") != NULL);
      ExP = (strstr(format,"xP") != NULL);
      ExQ = (strstr(format,"xQ") != NULL);
    } else {
      ExA = FALSE;
      ExB = FALSE;
      ExR = FALSE;
      ExS = FALSE;
      ExP = FALSE;
      ExQ = FALSE;
    }

    if (strstr(format,"Ax") != NULL) {
      AxA = (strstr(format,"xA") != NULL);
      AxB = (strstr(format,"xB") != NULL);
      AxR = (strstr(format,"xR") != NULL);
      AxS = (strstr(format,"xS") != NULL);
      AxP = (strstr(format,"xP") != NULL);
      AxQ = (strstr(format,"xQ") != NULL);
    } else {
      AxA = FALSE;
      AxB = FALSE;
      AxR = FALSE;
      AxS = FALSE;
      AxP = FALSE;
      AxQ = FALSE;
    }

    A2B = (strstr(format,"A2B") != NULL);
    R2S = (strstr(format,"R2S") != NULL);
    R2P = (strstr(format,"R2P") != NULL);
    R2Q = (strstr(format,"R2Q") != NULL);
    S2R = (strstr(format,"S2R") != NULL);
    S2P = (strstr(format,"S2P") != NULL);
    S2Q = (strstr(format,"S2Q") != NULL);
    P2Q = (strstr(format,"P2Q") != NULL);
    P2R = (strstr(format,"P2R") != NULL);
    Q2P = (strstr(format,"Q2P") != NULL);

    if (ExA) key = key + 1;
    if (AxA) key = key + 2;
    if (ExB) key = key + 4;
    if (AxB) key = key + 8;
    if (ExR) key = key + 16;
    if (AxR) key = key + 32;
    if (ExS) key = key + 64;
    if (AxS) key = key + 128;
    if (ExP) key = key + 256;
    if (AxP) key = key + 512;
    if (ExQ) key = key + 1024;
    if (AxQ) key = key + 2048;

    if (A2B) key = key + 4096;
    if (R2S) key = key + 8192;
    if (R2P) key = key + 16384;
    if (R2Q) key = key + 32768;
    if (S2R) key = key + 65536;
    if (S2P) key = key + 131072;
    if (S2Q) key = key + 262144;
    if (P2Q) key = key + 524288;
    if (P2R) key = key + 1048576;
    if (Q2P) key = key + 2097152;

    h = Bdd_RelOpSimple(f,format,FALSE);

  }

  if (!firsttime) {

    /* IS SOME NORMALIZATION POSSIBLE ??? */

    /* LOOKUP IN CACHE TABLE */

    bddRelOpSimpleCache.search++;

    /* NAIVE HASH FUNCTION */
    /*
    p = &bddRelOpSimpleCache.table[(
          ((uintptr_t) key << 0) +
          ((uintptr_t) f.p << 0)
		    ) % bddRelOpSimpleCache.size];
    */

#if UINTPTR_MAX == 0xffffffffffffffff

    /* MURMUR3 (A KIND OF) 64-bit HASH FUNCTION */
    /* https://code.google.com/p/smhasher/ */
    {
    const uint64_t C1 = 0x87c37b91114253d5L;
    const uint64_t C2 = 0x4cf5ad432745937fL;
    uintptr_t k1,k;
    k = key;
    k1 = (uintptr_t) f.p;
    k1 *= C1;
    k1 = (k1 << 31) | (k1 >> 33);
    k1 *= C2;
    k ^= k1;
    k = (k << 27) | (k >> 37);
    k = k*5+0x52dce729;
    k ^= k >> 33;
    k *= 0xff51afd7ed558ccdL;
    k ^= k >> 33;
    k *= 0xc4ceb9fe1a85ec53L;
    k ^= k >> 33;
    p = &bddRelOpSimpleCache.table[k & bddRelOpSimpleCache.size];
    }

#else

    /* MURMUR3 (A KIND OF) 32-bit HASH FUNCTION */
    /* https://code.google.com/p/smhasher/ */
    {
    const uint32_t C1 = 0xcc9e2d51;
    const uint32_t C2 = 0x1b873593;
    uintptr_t k1,k;
    k = key;
    k1 = (uintptr_t) f.p;
    k1 *= C1;
    k1 = (k1 << 15) | (k1 >> 17);
    k1 *= C2;
    k ^= k1;
    k = (k << 13) | (k >> 19);
    k = k*5+0xe6546b64;
    k ^= k >> 16;
    k *= 0x85ebca6b;
    k ^= k >> 13;
    k *= 0xc2b2ae35;
    k ^= k >> 16;
    p = &bddRelOpSimpleCache.table[k & bddRelOpSimpleCache.size];
    }

#endif

    if ((p->key == key) &&
        Bdd_isEqv(p->f,f))
    {
      bddRelOpSimpleCache.find++;
      h = p->r;
      Bdd_Refresh(h);
      return h;
    }

    /* DETERMINING THE SUCCESSORS AND TOP VARIABLE */

    f0 = Bdd_TransferMark(Bdd_GetElse(f), f.mark);
    f1 = Bdd_TransferMark(Bdd_GetThen(f), f.mark);
    zvar = Bdd_GetVariable(f);
    zchar = Bdd_GetVariableChar(f);

    /* DO RECURSIVE CALLS */

    h0 = Bdd_RelOpSimple(f0, format, firsttime);
    h1 = Bdd_RelOpSimple(f1, format, firsttime);

    /* COMPOSE PARTIAL RESULTS */
    switch (zchar) {

    case 'a':
    {
      done = FALSE;

      if (ExA) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxA) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (A2B) {
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "ab", thus b=a+1 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'b':
    {
      done = FALSE;

      if (ExB) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxB) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'r':
    {
      done = FALSE;

      if (ExR) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxR) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (R2S) {
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus s=r+1 */
        done = TRUE;
      }

      if (R2P) {
        h = Bdd_ITE(Bdd_Terminal(zvar+2), h1, h0); /* "rspq", thus p=r+2 */
        done = TRUE;
      }

      if (R2Q) {
        h = Bdd_ITE(Bdd_Terminal(zvar+3), h1, h0); /* "rspq", thus q=r+3 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 's':
    {
      done = FALSE;

      if (ExS) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxS) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (S2R) {
        h = Bdd_ITE(Bdd_Terminal(zvar-1), h1, h0); /* "rspq", thus r=s-1 */
        done = TRUE;
      }

      if (S2P) {
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus p=s+1 */
        done = TRUE;
      }

      if (S2Q) {
        h = Bdd_ITE(Bdd_Terminal(zvar+2), h1, h0); /* "rspq", thus q=s+2 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'p':
    {
      done = FALSE;

      if (ExP) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxP) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (P2Q) {
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus q=p+1 */
        done = TRUE;
      }

      if (P2R) {
        h = Bdd_ITE(Bdd_Terminal(zvar-2), h1, h0); /* "rspq", thus r=p-2 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'q':
    {
      done = FALSE;

      if (ExQ) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxQ) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (Q2P) {
        h = Bdd_ITE(Bdd_Terminal(zvar-1), h1, h0); /* "rspq", thus p=q-1 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    } /* switch zchar */

    /* STORE IN CACHE TABLE */

    if (p->key) bddRelOpSimpleCache.overwrite++;
    p->f = f;
    p->r = h;
    p->key = key;

  }

  return h;
}

/**Function****************************************************************
  Synopsis    []
  Description [Ex, Ax, R2S, and S2R can have a process specified.
               Symbol * means all processes.
               Symbol ^ means all other processes.
               Examples: "Ex xR<%s>", "S2R^<%s><%s>" ]
  SideEffects [NOT REENTRANT!]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_RelOpComplex(Bdd_Edge f, Est_String format, Est_Boolean firsttime)
{
  BddRelOpComplexCache *p; /* MUST NOT BE STATIC */
  Bdd_Edge f0, f1, h0, h1; /* MUST NOT BE STATIC */
  Bdd_Variable zvar; /* MUST NOT BE STATIC */
  char zchar; /* MUST NOT BE STATIC */
  Est_String pname; /* MUST NOT BE STATIC */

  Est_Boolean done; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */
  Bdd_Edge h; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */
  Est_String zname; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */
  Est_String t; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */
  int n; /* COULD BE STATIC TO MINIMIZE STACK PROBLEMS ? */

  static Est_Boolean ExA;  /* MUST BE STATIC */
  static Est_Boolean AxA;  /* MUST BE STATIC */
  static Est_String ExR;   /* MUST BE STATIC */
  static Est_String AxR;   /* MUST BE STATIC */
  static Est_String ExS;   /* MUST BE STATIC */
  static Est_String AxS;   /* MUST BE STATIC */
  static Est_String ExP;   /* MUST BE STATIC */
  static Est_String AxP;   /* MUST BE STATIC */
  static Est_String ExQ;   /* MUST BE STATIC */
  static Est_String AxQ;   /* MUST BE STATIC */

  static Est_String R2S;   /* MUST BE STATIC */
  static Est_String S2R;   /* MUST BE STATIC */

  static Est_Boolean R2P;   /* MUST BE STATIC */
  static Est_Boolean R2Q;   /* MUST BE STATIC */
  static Est_Boolean S2P;   /* MUST BE STATIC */
  static Est_Boolean S2Q;   /* MUST BE STATIC */
  static Est_Boolean P2Q;   /* MUST BE STATIC */
  static Est_Boolean Q2P;   /* MUST BE STATIC */

  static unsigned int key;  /* CACHE TABLE - MUST BE STATIC */

  if (Bdd_isTerminal(f)) return f;

  /* IF FIRSTIME, PARSING FORMAT STRING */

  if (firsttime) {

    /* DEBUGGING */
    /*
    if (!BddisOK(f)) {
      fprintf(stderr,"ERROR (Bdd_RelOpComplex): Bad f!\n");
      free((void *)1);
      exit(1);
    }
    */

    bddRelOpComplexCache.num++;

    key = 0;                              /* 0 * 2^30 */

    if (strstr(format,"Ex") != NULL) {

      ExA = (strstr(format,"xA") != NULL);

      if ((t = strstr(format,"xR"))) {
        n = strcspn(&t[2]," ");
        if (!(ExR = (Est_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memmove(ExR,&t[2],n);
        ExR[n] = 0;

        /*
        ExR = (Est_String) strtok(strdup(&t[2]), " ");
        */

      } else {
        ExR = NULL;
      }

      if ((t = strstr(format,"xS"))) {
        n = strcspn(&t[2]," ");
        if (!(ExS = (Est_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memmove(ExS,&t[2],n);
        ExS[n] = 0;

        /*
        ExS = (Est_String) strtok(strdup(&t[2]), " ");
        */

      } else {
        ExS = NULL;
      }

      if ((t = strstr(format,"xP"))) {
        n = strcspn(&t[2]," ");
        if (!(ExP = (Est_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memmove(ExP,&t[2],n);
        ExP[n] = 0;

        /*
        ExP = (Est_String) strtok(strdup(&t[2]), " ");
        */

      } else {
        ExP = NULL;
      }

      if ((t = strstr(format,"xQ"))) {
        n = strcspn(&t[2]," ");
        if (!(ExQ = (Est_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memmove(ExQ,&t[2],n);
        ExQ[n] = 0;

        /*
        ExQ = (Est_String) strtok(strdup(&t[2]), " ");
        */

      } else {
        ExQ = NULL;
      }

    } else {
      ExA = FALSE;
      ExR = NULL;
      ExS = NULL;
      ExP = NULL;
      ExQ = NULL;
    }

    if (strstr(format,"Ax") != NULL) {
      AxA = (strstr(format,"xA") != NULL);

      if ((t = strstr(format,"xR"))) {
        n = strcspn(&t[2]," ");
        if (!(AxR = (Est_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memmove(AxR,&t[2],n);
        AxR[n] = 0;

        /*
        AxR = (Est_String) strtok(strdup(&t[2]), " ");
        */

      } else {
        AxR = NULL;
      }

      if ((t = strstr(format,"xS"))) {
        n = strcspn(&t[2]," ");
        if (!(AxS = (Est_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memmove(AxS,&t[2],n);
        AxS[n] = 0;

        /*
        AxS = (Est_String) strtok(strdup(&t[2]), " ");
        */

      } else {
        AxS = NULL;
      }

      if ((t = strstr(format,"xP"))) {
        n = strcspn(&t[2]," ");
        if (!(AxP = (Est_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memmove(AxP,&t[2],n);
        AxP[n] = 0;

        /*
        AxP = (Est_String) strtok(strdup(&t[2]), " ");
        */

      } else {
        AxP = NULL;
      }

      if ((t = strstr(format,"xQ"))) {
        n = strcspn(&t[2]," ");
        if (!(AxQ = (Est_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memmove(AxQ,&t[2],n);
        AxQ[n] = 0;

        /*
        AxQ = (Est_String) strtok(strdup(&t[2]), " ");
        */

      } else {
        AxQ = NULL;
      }

    } else {
      AxA = FALSE;
      AxR = NULL;
      AxS = NULL;
      AxP = NULL;
      AxQ = NULL;
    }

    if ((t = strstr(format,"R2S"))) {
      n = strcspn(&t[3]," ");
      if (!(R2S = (Est_String) malloc(n+1))) {
        fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
        exit(1);
      }
      memmove(R2S,&t[3],n);
      R2S[n] = 0;
    } else {
      R2S = NULL;
    }

    if ((t = strstr(format,"S2R"))) {
      n = strcspn(&t[3]," ");
      if (!(S2R = (Est_String) malloc(n+1))) {
        fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
        exit(1);
      }
      memmove(S2R,&t[3],n);
      S2R[n] = 0;

    } else {
      S2R = NULL;
    }

    R2P = (strstr(format,"R2P") != NULL);
    R2Q = (strstr(format,"R2Q") != NULL);
    S2P = (strstr(format,"S2P") != NULL);
    S2Q = (strstr(format,"S2Q") != NULL);
    P2Q = (strstr(format,"P2Q") != NULL);
    Q2P = (strstr(format,"Q2P") != NULL);

    if (ExA) key = key + 1;
    if (AxA) key = key + 2;
    if (ExR) key = key + 4;
    if (AxR) key = key + 8;
    if (ExS) key = key + 16;
    if (AxS) key = key + 32;
    if (ExP) key = key + 64;
    if (AxP) key = key + 128;
    if (ExQ) key = key + 256;
    if (AxQ) key = key + 512;

    if (R2S) key = key + 1024;
    if (R2P) key = key + 2048;
    if (R2Q) key = key + 4096;
    if (S2R) key = key + 8192;
    if (S2P) key = key + 16384;
    if (S2Q) key = key + 32768;
    if (P2Q) key = key + 65536;
    if (Q2P) key = key + 131072;

    if (ExR && (ExR[0]=='*')) {
      key = key + 4194304;
    }

    if (AxR && (AxR[0]=='*')) {
      key = key + 4194304;
    }

    if (ExS && (ExS[0]=='*')) {
      key = key + 8388608;
    }

    if (AxS && (AxS[0]=='*')) {
      key = key + 8388608;
    }

    if (ExP && (ExP[0]=='*')) {
      key = key + 16777216;
    }

    if (AxP && (AxP[0]=='*')) {
      key = key + 16777216;
    }

    if (ExQ && (ExQ[0]=='*')) {
      key = key + 33554432;
    }

    if (AxQ && (AxQ[0]=='*')) {
      key = key + 33554432;
    }

    h = Bdd_RelOpComplex(f,format,FALSE);

    if (ExR) free(ExR);
    if (AxR) free(AxR);
    if (ExS) free(ExS);
    if (AxS) free(AxS);
    if (ExP) free(ExP);
    if (AxP) free(AxP);
    if (ExQ) free(ExQ);
    if (AxQ) free(AxQ);
  }

  if (!firsttime) {

    /* IS SOME NORMALIZATION POSSIBLE ??? */

    /* LOOKUP IN CACHE TABLE */

    bddRelOpComplexCache.search++;

    /* NAIVE HASH FUNCTION */
    /*
    p = &bddRelOpComplexCache.table[(
          ((uintptr_t) key << 0) +
          ((uintptr_t) f.p << 0)
		    ) % bddRelOpComplexCache.size];
    */

#if UINTPTR_MAX == 0xffffffffffffffff

    /* MURMUR3 (A KIND OF) 64-bit HASH FUNCTION */
    /* https://code.google.com/p/smhasher/ */
    {
    const uint64_t C1 = 0x87c37b91114253d5L;
    const uint64_t C2 = 0x4cf5ad432745937fL;
    uintptr_t k1,k;
    k = key;
    k1 = (uintptr_t) f.p;
    k1 *= C1;
    k1 = (k1 << 31) | (k1 >> 33);
    k1 *= C2;
    k ^= k1;
    k = (k << 27) | (k >> 37);
    k = k*5+0x52dce729;
    k ^= k >> 33;
    k *= 0xff51afd7ed558ccdL;
    k ^= k >> 33;
    k *= 0xc4ceb9fe1a85ec53L;
    k ^= k >> 33;
    p = &bddRelOpComplexCache.table[k & bddRelOpComplexCache.size];
    }

#else

    /* MURMUR3 (A KIND OF) 32-bit HASH FUNCTION */
    /* https://code.google.com/p/smhasher/ */
    {
    const uint32_t C1 = 0xcc9e2d51;
    const uint32_t C2 = 0x1b873593;
    uintptr_t k1,k;
    k = key;
    k1 = (uintptr_t) f.p;
    k1 *= C1;
    k1 = (k1 << 15) | (k1 >> 17);
    k1 *= C2;
    k ^= k1;
    k = (k << 13) | (k >> 19);
    k = k*5+0xe6546b64;
    k ^= k >> 16;
    k *= 0x85ebca6b;
    k ^= k >> 13;
    k *= 0xc2b2ae35;
    k ^= k >> 16;
    p = &bddRelOpComplexCache.table[k & bddRelOpComplexCache.size];
    }

#endif

    if ((p->key == key) &&
        Bdd_isEqv(p->f,f) &&
        !strcmp(p->name,format))
    {
      bddRelOpComplexCache.find++;
      h = p->r;
      Bdd_Refresh(h);
      return h;
    }

    /* DETERMINING THE SUCCESSORS, TOP VARIABLE, AND THE NAME OF THE PROCESS */

    f0 = Bdd_TransferMark(Bdd_GetElse(f), f.mark);
    f1 = Bdd_TransferMark(Bdd_GetThen(f), f.mark);
    zvar = Bdd_GetVariable(f);
    zname = Bdd_GetIthVariableName(zvar);
    zchar = zname[0];
    if (zchar != 'a') {
      pname = &zname[1+strspn(&zname[1],"0123456789")];
    }

    /* DO RECURSIVE CALLS */

    h0 = Bdd_RelOpComplex(f0, format, firsttime);
    h1 = Bdd_RelOpComplex(f1, format, firsttime);

    /* COMPOSE PARTIAL RESULTS */
    switch (zchar) {

    case 'a':
    {
      done = FALSE;

      if (ExA) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxA) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'r':
    {
      done = FALSE;

      if (ExR && ((ExR[0]=='*') || (ExR[0]!='^' && strstr(ExR,pname)) || (ExR[0]=='^' && !strstr(ExR,pname)))) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxR && ((AxR[0]=='*') || (AxR[0]!='^' && strstr(AxR,pname)) || (AxR[0]=='^' && !strstr(AxR,pname)))) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (R2S && ((R2S[0]=='*') || (R2S[0]!='^' && strstr(R2S,pname)) || (R2S[0]=='^' && !strstr(R2S,pname)))) {
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus s=r+1 */
        done = TRUE;
      }

      if (R2P) {
        h = Bdd_ITE(Bdd_Terminal(zvar+2), h1, h0); /* "rspq", thus p=r+2 */
        done = TRUE;
      }

      if (R2Q) {
        h = Bdd_ITE(Bdd_Terminal(zvar+3), h1, h0); /* "rspq", thus q=r+3 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 's':
    {
      done = FALSE;

      if (ExS && ((ExS[0]=='*') || (ExS[0]!='^' && strstr(ExS,pname)) || (ExS[0]=='^' && !strstr(ExS,pname)))) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxS && ((AxS[0]=='*') || (AxS[0]!='^' && strstr(AxS,pname)) || (AxS[0]=='^' && !strstr(AxS,pname)))) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (S2R && ((S2R[0]=='*') || (S2R[0]!='^' && strstr(S2R,pname)) || (S2R[0]=='^' && !strstr(S2R,pname)))) {
        h = Bdd_ITE(Bdd_Terminal(zvar-1), h1, h0); /* "rspq", thus r=s-1 */
        done = TRUE;
      }

      if (S2P) {
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus p=s+1 */
        done = TRUE;
      }

      if (S2Q) {
        h = Bdd_ITE(Bdd_Terminal(zvar+2), h1, h0); /* "rspq", thus q=s+2 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'p':
    {
      done = FALSE;

      if (ExP && ((ExP[0]=='*') || (ExP[0]!='^' && strstr(ExP,pname)) || (ExP[0]=='^' && !strstr(ExP,pname)))) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxP && ((AxP[0]=='*') || (AxP[0]!='^' && strstr(AxP,pname)) || (AxP[0]=='^' && !strstr(AxP,pname)))) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (P2Q) {
        h = Bdd_ITE(Bdd_Terminal(zvar+1), h1, h0); /* "rspq", thus q=p+1 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    case 'q':
    {
      done = FALSE;

      if (ExQ && ((ExQ[0]=='*') || (ExQ[0]!='^' && strstr(ExQ,pname)) || (ExQ[0]=='^' && !strstr(ExQ,pname)))) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxQ && ((AxQ[0]=='*') || (AxQ[0]!='^' && strstr(AxQ,pname)) || (AxQ[0]=='^' && !strstr(AxQ,pname)))) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (Q2P) {
        h = Bdd_ITE(Bdd_Terminal(zvar-1), h1, h0); /* "rspq", thus p=q-1 */
        done = TRUE;
      }

      if (!done) {
        h = Bdd_ITE(Bdd_Terminal(zvar), h1, h0);
      }

      break;
    }

    } /* switch zchar */

    /* STORE IN CACHE TABLE */

    if (p->key) {
      bddRelOpComplexCache.overwrite++;
      free(p->name);
    }
    p->f = f;
    p->r = h;
    p->key = key;
    p->name = strdup(format);

  }

  return h;
}

/**Function****************************************************************
  Synopsis    [NewState]
  Description ['s' variables replaces with 'r'. Performs existential
               quantification over all other variables.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_NewState(Bdd_Edge f)
{
  return Bdd_RelOpSimple(f,"Ex xR xA xP xQ S2R",TRUE);
}

/**Function****************************************************************
  Synopsis    [NewState1]
  Description ['s' variables replaces with 'r'.
               Performs existential quatification over 'r' variables
               of the given process and existential quatification over
               other variables.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_NewState1(Bdd_Edge f, Est_String name)
{
  Est_String s;
  Bdd_Edge r;

  s = (Est_String) malloc(255);
  sprintf(s,"Ex xA xR<%s> xP* xQ* S2R*",name);
  r = Bdd_RelOpComplex(f,s,TRUE);
  free(s);
  return r;
}

/**Function****************************************************************
  Synopsis    []
  Description ['s' variables replaces with 'r'.
               Performs existential quantification over 'r' variables
               of the given two processes and existential quantification
               over other variables.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_NewState2(Bdd_Edge f, Est_String name1, Est_String name2)
{
  Est_String s;
  Bdd_Edge r;

  s = (Est_String) malloc(255);
  sprintf(s,"Ex xA xR<%s><%s> xP* xQ* S2R*",name1,name2);
  r = Bdd_RelOpComplex(f,s,TRUE);
  free(s);
  return r;
}

/**Function****************************************************************
  Synopsis    []
  Description [Calculate f->g. Performs universal quantification over
               'a', 's' and 'q'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_NewEq(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#IMPL Ax xA xS xQ",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Add stab, i.e. for each positive 'r' variable adds
               positive 's' variable and for each negative 'r'
               variable adds negative 's' variable. IT USES stablist
               WHICH MUST BE A PRODUCT OF ALL r VARIABLES WHICH
               NEED THE ADDITION OF CORRESPONDING 's'!]
  SideEffects ['r' and corresponding 's' must be consecutive]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_StabRS(Bdd_Edge f, Bdd_Edge stablist)
{
  BddStabRSCache *p;
  Bdd_Edge f0, f1, h0, h1, h;
  Bdd_Variable fvar,svar;

  /* DEBUGGING */
  /*
  if (Bdd_isEqv(stablist,bdd_termFalse)) {
    printf("ERROR: Bdd_StabRS get stablist == bdd_termFalse\n");
    return bdd_termNull;
  }
  */

  if (Bdd_isEqv(stablist,bdd_termTrue)) {
    return f;
  }
  if (Bdd_isEqv(f,bdd_termFalse)) {
    return f;
  }

  /* IS SOME NORMALIZATION POSSIBLE ??? */

  /* LOOKUP IN CACHE TABLE */

  bddStabRSCache.search++;

  /* NAIVE HASH FUNCTION */
  /*
  p = &bddStabRSCache.table[(
        ((uintptr_t) f.p << 0) +
        ((uintptr_t) stablist.p << 3)
      ) % bddStabRSCache.size];
  */

#if UINTPTR_MAX == 0xffffffffffffffff

  /* MURMUR3 (A KIND OF) 64-bit HASH FUNCTION */
  /* https://code.google.com/p/smhasher/ */
  {
  const uint64_t C1 = 0x87c37b91114253d5L;
  const uint64_t C2 = 0x4cf5ad432745937fL;
  uintptr_t k1,k;
  k = (uintptr_t) f.p;
  k1 = (uintptr_t) stablist.p;
  k1 *= C1;
  k1 = (k1 << 31) | (k1 >> 33);
  k1 *= C2;
  k ^= k1;
  k = (k << 27) | (k >> 37);
  k = k*5+0x52dce729;
  k ^= k >> 33;
  k *= 0xff51afd7ed558ccdL;
  k ^= k >> 33;
  k *= 0xc4ceb9fe1a85ec53L;
  k ^= k >> 33;
  p = &bddStabRSCache.table[k & bddStabRSCache.size];
  }

#else

  /* MURMUR3 (A KIND OF) 32-bit HASH FUNCTION */
  /* https://code.google.com/p/smhasher/ */
  {
  const uint32_t C1 = 0xcc9e2d51;
  const uint32_t C2 = 0x1b873593;
  uintptr_t k1,k;
  k = (uintptr_t) f.p;
  k1 = (uintptr_t) stablist.p;
  k1 *= C1;
  k1 = (k1 << 15) | (k1 >> 17);
  k1 *= C2;
  k ^= k1;
  k = (k << 13) | (k >> 19);
  k = k*5+0xe6546b64;
  k ^= k >> 16;
  k *= 0x85ebca6b;
  k ^= k >> 13;
  k *= 0xc2b2ae35;
  k ^= k >> 16;
  p = &bddStabRSCache.table[k & bddStabRSCache.size];
  }

#endif

  if (!Bdd_isNull(p->r) && Bdd_isEqv(p->f,f) && Bdd_isEqv(p->stablist,stablist))
  {
    bddStabRSCache.find++;
    h = p->r;
    Bdd_Refresh(h);
    return h;
  }

  /* DETERMINING THE SUCCESSORS AND TOP VARIABLE */
  if (Bdd_isEqv(f,bdd_termTrue)) {
    f0 = bdd_termTrue;
    f1 = bdd_termTrue;
  } else {
    f0 = Bdd_TransferMark(Bdd_GetElse(f), f.mark);
    f1 = Bdd_TransferMark(Bdd_GetThen(f), f.mark);
  }
  fvar = Bdd_GetVariable(f);
  svar = Bdd_GetVariable(stablist);

  if (Bdd_IsVariableSmaller(svar,fvar)) {
    h = Bdd_StabRS(f,Bdd_GetThen(stablist));
    h0 = Bdd_ITE(Bdd_Terminal(svar+1),bdd_termFalse,h); /* add negative 's' */
    h1 = Bdd_ITE(Bdd_Terminal(svar+1),h,bdd_termFalse); /* add positive 's' */
    fvar = svar;
  } else if (Bdd_IsVariableSmaller(fvar,svar)) {
    h0 = Bdd_StabRS(f0,stablist);
    h1 = Bdd_StabRS(f1,stablist);
  } else {
    h0 = Bdd_ITE(Bdd_Terminal(fvar+1),bdd_termFalse,Bdd_StabRS(f0,Bdd_GetThen(stablist))); /* add negative 's' */
    h1 = Bdd_ITE(Bdd_Terminal(fvar+1),Bdd_StabRS(f1,Bdd_GetThen(stablist)),bdd_termFalse); /* add positive 's' */
  }

  h = Bdd_ITE(Bdd_Terminal(fvar), h1, h0);

  /* STORE IN CACHE TABLE */

  if (!Bdd_isNull(p->r)) {
    bddStabRSCache.overwrite++;
  }
  p->f = f;
  p->stablist = stablist;
  p->r = h;

  return h;
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/
