/**CFile*******************************************************************
  PackageName [bdd]
  Synopsis    [Package 'bdd' enable symbolic computations by representing
               Boolean functions with ROBDDs.]

  FileName    [bddUtil.c]
  Revision    [$Revision: 113 $]
  Date        [$Date: 2016-03-08 22:31:16 +0100 (tor, 08 mar 2016) $]
  Authors     [Robert Meolic (robert.meolic@um.si),
               Ales Casar (ales@homemade.net)]
  Description [File bddUtil.c contains useful functions.]
  SeeAlso     [bdd.h, bddInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 2003, 2016
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include "bddInt.h"

#define oAND 1
#define oOR 2
#define oEXOR 3

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static Bdd_Edge Izraz(Est_String s, int *i, Est_String *ch);

static void NextCh(Est_String s, int *i, Est_String *ch);

static Bdd_Edge Izrazi(Est_String s, int *i, Est_String *ch, Bdd_Edge g,
                        int op);

static int Op(Est_String s, int *i, Est_String *ch);

static Est_Boolean DoberZnak(char c);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function Bdd_ExtractBranch]
  Description [Returns sub-BDD with only one (left-most) one-path.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_ExtractBranch(Bdd_Edge f)
{
  if (Bdd_isTerminal(f)) return f;
  if (!f.mark) {
    if (!Bdd_isEqv(Bdd_GetElse(f),bdd_termFalse)) {
      return BddFoaNode(Bdd_GetVariable(f),Bdd_ExtractBranch(Bdd_GetElse(f)),bdd_termFalse);
    } else {
      return BddFoaNode(Bdd_GetVariable(f),bdd_termFalse,Bdd_ExtractBranch(Bdd_GetThen(f)));
    }
  } else {
    if (!Bdd_isEqv(Bdd_GetElse(f),bdd_termTrue)) {
      return BddFoaNode(Bdd_GetVariable(f),Bdd_ExtractBranch(Bdd_NOT(Bdd_GetElse(f))),bdd_termFalse);
    } else {
      return BddFoaNode(Bdd_GetVariable(f),bdd_termFalse,Bdd_ExtractBranch(Bdd_NOT(Bdd_GetThen(f))));
    }
  }

  return bdd_termNull;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_Expand]
  Description [Bdd_Expand expands Boolean function in such a way, that
               it substitute all variables with corresponding functions.]
  SideEffects [In some cases, some subfunctions are also expanded]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_Expand(Est_String fname, Bdd_Edge f)
{
  Est_String topname;
  Bdd_Edge g;

  topname = Bdd_GetVariableName(f);

  /*
  printf("START Bdd_Expand %s\n",fname);
  Bdd_WriteFunction(f);
  Bdd_WriteBDD(f);
  */

  if (Bdd_isTerminal(f)) {
    return f;
  }

  if (!Bdd_FindFormula(topname,&g)) {
    return f;
  };

  if (Bdd_IsSmaller(f,g)) {
    /*
    printf("topF <<\n");
    */
    f = Bdd_Compose(f,Bdd_GetVariable(f),g); /* SUBSTITUTE SUBFUNCTION ... */
    Bdd_SaveFormula(NULL,fname,f);           /* SAVE RESULT ... */
    f = Bdd_Expand(fname,f);                 /* AND THEN TRY EXPANDING f AGAIN */
  }

  else if (Bdd_IsSmaller(g,f)) {
    /*
    printf("topF >>\n");
    */
    g = Bdd_Expand(topname,g); /* EXPAND SUBFUNCTION FIRST ... */
    f = Bdd_Expand(fname,f);   /* AND THEN TRY EXPANDING f AGAIN */
  }

  else {
    /*
    printf("topF ==\n");
    */
  }

  return f;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_Expr]
  Description []
  SideEffects [DEPRECATED. Use Bdd_Eval1 instead of this.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_Expr(Est_String s)
{
  Bdd_Edge pom;
  Est_String ch;
  int i;

  ch = (Est_String) malloc(255);

  i = 0;
  NextCh(s,&i,&ch);
  pom = Izraz(s,&i,&ch);

  free(ch);
  return pom;
}

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function Izraz]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
Izraz(Est_String s, int *i, Est_String *ch)
{
  Bdd_Edge f;
  int op;

  if (!strcmp(*ch,"(")) {
    NextCh(s,i,ch);

    if (!strcmp(*ch, "NOT")) {
      NextCh(s,i,ch);
      f = Bdd_NOT(Izraz(s,i,ch));
    } else {
      if ((op = Op(s,i,ch))) {
	f = Izrazi(s, i, ch, Izraz(s,i,ch), op);
      } else {
	f = Izraz(s,i,ch);
      }
    }

    if (!strcmp(*ch, ")")) NextCh(s,i,ch);

    return f;
  }

  if (DoberZnak(*ch[0])) {
    if (!Bdd_FindFormula(*ch, &f)) f = Bdd_FoaTerminal(*ch);
    NextCh(s,i,ch);
    return f;
  }

  return bdd_termNull;
}

/**Function****************************************************************
  Synopsis    [Function Izrazi]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
Izrazi(Est_String s, int *i, Est_String *ch, Bdd_Edge g, int op)
{
  Bdd_Edge f, h;

  h = Izraz(s,i,ch);
  if (!Bdd_isNull(h)) {
    switch (op) {
	    case oAND:  f = Bdd_ITE(g, h, bdd_termFalse);
	                break;
	    case oOR:  f = Bdd_ITE(g, bdd_termTrue, h);
	                break;
	    case oEXOR:  f = Bdd_ITE(g, Bdd_NOT(h), h);
    }
    return Izrazi(s, i, ch, f, op);
  }
  return g;
}

/**Function****************************************************************
  Synopsis    [Function NextCh]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
NextCh(Est_String s, int *i, Est_String *ch)
{
  char c;
  int j;

    while (s[*i] == ' ' || s[*i] == '\t') (*i)++;
    j = *i;
    c = s[*i];

    if (c == '(') {
      (*i)++;
      strcpy(*ch,"(");
      return;
    }

    if (c == ')') {
      (*i)++;
      strcpy(*ch,")");
      return;
    }

    if (DoberZnak(c)) {
      (*i)++;
      while (DoberZnak(s[*i])) (*i)++;
      c = s[*i];

      if ( (c == ' ') || (c == '\t') || (c == '(') || (c == ')') ) {
        strncpy(*ch, &(s[j]), *i-j);
        (*ch)[*i-j] = 0;
        return;
      }
    }

    strcpy(*ch,"");
    return;
}

/**Function****************************************************************
  Synopsis    [Function Op]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
Op(Est_String s, int *i, Est_String *ch)
{
  if (!strcmp(*ch, "AND")) {
    NextCh(s,i,ch);
    return oAND;
  }

  if (!strcmp(*ch, "OR")) {
    NextCh(s,i,ch);
    return oOR;
  }

  if (!strcmp(*ch, "EXOR")) {
    NextCh(s,i,ch);
    return oEXOR;
  }

  return 0;
}

/**Function****************************************************************
  Synopsis    [Function DoberZnak]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Est_Boolean
DoberZnak(char c)
{
  return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') ||
         (c >= '0' && c <= '9') || c == '_' || c == '[' || c == ']' ||
         c == '$' || c == '.' || c == '&';
}

/**Function****************************************************************
  Synopsis    [Function UpCase]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Est_String
UpCase(Est_String s)
  {
    Est_String p;

    p = s;
    while (p[0])
      {
	p[0] = toupper(p[0]);
	p = &p[1];
      }
    return s;
  }
