/**CFile*******************************************************************
  PackageName [gui]
  Synopsis    [Package 'gui' is used to redirect output to GUI.]

  FileName    [guiTcl.c]
  Revision    [$Revision: 127 $]
  Date        [$Date: 2016-07-14 02:22:29 +0200 (čet, 14 jul 2016) $]
  Authors     [Robert Meolic (robert.meolic@um.si)]
  Description [File guiTcl.c contains definitions of Tcl commands, which
               can be used from Tcl interpreter (e.g. tclsh or wish).]
  SeeAlso     [gui.h, guiInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 2003, 2016
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include "guiInt.h"
#include "guiDecls.h"

#include <tcl.h>

extern GuiStubs guiStubs;

/* on tcl 8.3 use #define USECONST */
/* on tcl 8.4 use #define USECONST const*/
/* this is defined in Makefile */

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/* EXPORTED VARIABLES */

Gui_Output gui_outputStream = GUIERROR; /* redirecting output to socket */
int gui_outputLength = -1; /* counting bytes mechanism, a shared variable */

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static int GuiInitPkgCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int GuiExitPkgCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int GuiAboutPkgCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int GuiOutputCmd(ClientData clientData, Tcl_Interp *interp,
                        int argc, USECONST char **argv);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

int
Gui_Init(Tcl_Interp *interp)
{

#ifdef USE_TCL_STUBS
  if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
    return TCL_ERROR;
  }
#endif

  Tcl_LinkVar(interp, "GUIOUTPUTLENGTH", (char *) &gui_outputLength,
                     TCL_LINK_INT);
                     
  Tcl_CreateCommand(interp, "gui_initPkg", GuiInitPkgCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "gui_exitPkg", GuiExitPkgCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "gui_aboutPkg", GuiAboutPkgCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "gui_output", GuiOutputCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  return Tcl_PkgProvideEx(interp, "est-gui", "1.0", &guiStubs);
}

#ifdef __cplusplus
}
#endif

/**Function****************************************************************
  Synopsis    [Function Gui_Printf.]
  Description [Redirect printf to tcl.]
  SideEffects [DEPRECATED. Printf is now redirected to socket.]
  SeeAlso     []
  ************************************************************************/

int
Gui_Printf(char *format, ...)
{
  return 0;
}

/**Function****************************************************************
  Synopsis    [Function Gui_Update.]
  Description [Used to refresh GUI after printf.]
  SideEffects [Should be used only if client and server are in the same
               process, e.g. myinterface (server) + gui package (client)]
  SeeAlso     []
  ************************************************************************/
  
void
Gui_Update() {
  if (gui_outputLength == -2) return; /* remote server! */
  Tcl_DoOneEvent(0);
  do {} while (Tcl_DoOneEvent(TCL_DONT_WAIT));
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function GuiInitPkgCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
GuiInitPkgCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                  USECONST char **argv)
{

  Est_String s0,s1,s2;
  int mode = 0;

#ifdef USESOCKETS
#ifdef ESTWIN32

  SOCKET ConnectSocket = INVALID_SOCKET;
  struct addrinfo *result = NULL, *ptr = NULL, hints;
  WSADATA wsaData;
  int iResult;

#else

  int port;
  struct hostent *he;
  struct in_addr **addr_list;
  struct sockaddr_in server;
  int sock;

#endif
#endif

#ifdef USESOCKETS
  if (argc != 4) {
    Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
    return TCL_ERROR;
  }
  s0 = strdup(argv[1]);
  s1 = strdup(argv[2]);
  s2 = strdup(argv[3]);
  sscanf(s0,"%d",&mode);

#else
  if (argc != 1) {
    Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
    return TCL_ERROR;
  }
#endif


#ifdef USESOCKETS
#ifdef ESTWIN32

  /* Initialize Winsock */
  iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
  if (iResult != 0) {
    fprintf(stderr,"WSAStartup failed: %d\n", iResult);
    free(s1);
    free(s2);
    return TCL_ERROR;
  }

  ZeroMemory( &hints, sizeof(hints) );
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_protocol = IPPROTO_TCP;

  /* Resolve the server address and port */
  iResult = getaddrinfo(s1,s2,&hints,&result);
  if (iResult != 0) {
    fprintf(stderr,"getaddrinfo failed: %d\n", iResult);
    WSACleanup();
    free(s1);
    free(s2);
    return TCL_ERROR;
  }

  /* Attempt to connect to the first address returned by the call to getaddrinfo */
  ptr=result;

  /* Create a SOCKET for connecting to server */
  ConnectSocket = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);

  if (ConnectSocket == INVALID_SOCKET) {
    fprintf(stderr,"Error at socket(): %ld\n", WSAGetLastError());
    freeaddrinfo(result);
    WSACleanup();
    free(s1);
    free(s2);
    return TCL_ERROR;
  }

  /* Connect to server */
  iResult = connect(ConnectSocket, ptr->ai_addr, (int)ptr->ai_addrlen);
  if (iResult == SOCKET_ERROR) {
    closesocket(ConnectSocket);
    ConnectSocket = INVALID_SOCKET;
  }

  freeaddrinfo(result);

  if (ConnectSocket == INVALID_SOCKET) {
    fprintf(stderr,"Unable to connect to server!\n");
    WSACleanup();
    free(s1);
    free(s2);
    return TCL_ERROR;
  }

  gui_outputStream = ConnectSocket;

  if (gui_outputStream == GUIERROR) {
    fprintf(stderr,"Unable to connect to server!\n");
    WSACleanup();
    free(s1);
    free(s2);
    return TCL_ERROR;
  }
  
  free(s1);
  free(s2);

  /*
  iResult = send(gui_outputStream, sendbuf, (int) strlen(sendbuf), 0);
  if (iResult == SOCKET_ERROR) {
    fprintf(stderr,"send failed: %d\n", WSAGetLastError());
    closesocket(gui_outputStream);
    WSACleanup();
    return TCL_ERROR;
  }
  */

#else

  /* create client socket */
  sock = socket(AF_INET,SOCK_STREAM,0);
  if (sock == -1) {
    fprintf(stderr,"ERROR: create socket failed\n");
    free(s1);
    free(s2);
    return TCL_ERROR;
  }

  /* function hostname_to_ip - thanks to Silver Moon */
  /* http://www.binarytides.com/hostname-to-ip-address-c-sockets-linux/ */
  if ((he=gethostbyname(s1)) == NULL) {
    fprintf(stderr,"ERROR: gethostbyname failed\n");
    free(s1);
    free(s2);
    return TCL_ERROR;
  }
  addr_list = (struct in_addr **) he->h_addr_list;
  if (addr_list[0] == NULL) {
    fprintf(stderr,"ERROR: gethostbyname problem\n");
    free(s1);
    free(s2);
    return TCL_ERROR;
  }

  /* connect to server socket */
  server.sin_addr.s_addr = inet_addr(inet_ntoa(*addr_list[0]));
  server.sin_family = AF_INET;
  sscanf(s2,"%d",&port);
  server.sin_port = htons(port);
  if (connect(sock,(struct sockaddr *)&server,sizeof(server)) < 0 )
  {
    fprintf(stderr,"ERROR: connect failed\n");
    free(s1);
    free(s2);
    return TCL_ERROR;
  }

  if (!(gui_outputStream = fdopen(sock,"w"))) {
    fprintf(stderr,"ERROR: fdopen failed\n");
    free(s1);
    free(s2);
    return TCL_ERROR;
  }
  setbuf(gui_outputStream,NULL);

  free(s1);
  free(s2);

#endif
#endif

  if (mode == 0) {
    gui_outputLength = -2; /* remote server */
  } else if (mode == 1) {
    gui_outputLength = -1; /* do not use counting bytes mechanism */
  } else if (mode == 2) {
    gui_outputLength = 0; /* use counting bytes mechanism */
  }

  Gui_InitPkg();

#ifdef USESOCKETS
  if (gui_outputLength == -2) /* remote server */
  printf("EST_TCL_OK\n");
#endif

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function GuiExitPkgCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
GuiExitPkgCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                   USECONST char **argv)
{
  if (argc != 1) {
    Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
    return TCL_ERROR;
  }

  printf("Exit GUI package... ");
  Gui_ExitPkg();
  printf("OK\n");

#ifdef USESOCKETS
#ifdef ESTWIN32

  /* closesocket(); */
  WSACleanup();

#else

  fclose(gui_outputStream);

#endif
#endif

  gui_outputStream = GUIERROR;

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function GuiAboutPkgCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
GuiAboutPkgCmd(ClientData clientData, Tcl_Interp *interp, int argc,
              USECONST char **argv)
{
  if (argc != 1) {
    Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
    return TCL_ERROR;
  }

  Gui_AboutPkg();

#ifdef USESOCKETS
  if (gui_outputLength == -2) /* remote server */
  printf("EST_TCL_OK\n");
#endif

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function GuiOutputCmd.]
  Description [Redirect tcl output to printf.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
GuiOutputCmd(ClientData clientData, Tcl_Interp *interp, int argc,
              USECONST char **argv)
{
  Est_String s;

  if (argc != 2) {
    Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
    return TCL_ERROR;
  }

  s = strdup(argv[1]);
  printf("%s",s);

  free(s);

#ifdef USESOCKETS
  if (gui_outputLength == -2) /* remote server */
  printf("EST_TCL_OK\n");
#endif

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}
