/**CFile*******************************************************************
  PackageName [pa]
  Synopsis    [Package 'pa' is a framework for implementation of
               process algebrae similar to Milner's CCS and Hoare's CSP.]

  FileName    [paMain.c]
  Revision    [$Revision: 126 $]
  Date        [$Date: 2016-07-12 20:51:30 +0200 (tor, 12 jul 2016) $]
  Authors     [Robert Meolic (robert.meolic@um.si)]
  Description [File paMain.c contains main functions.]
  SeeAlso     [pa.h, paInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 2003, 2016
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include "paInt.h"

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/* EXPORTED VARIABLES */

int pa_status = 0;                           /* initialization status */
unsigned short int pa_variables = 0;         /* number of variables */
int pa_sorts = 0;                            /* number of sorts */
int pa_processes = 0;                        /* number of processes */
int pa_compositions = 0;                     /* number of compositions */
Pa_Variable *pa_variableTable = NULL;        /* variable table */
Pa_Sort *pa_sortTable = NULL;                /* table of sorts */
Pa_Process *pa_processTable = NULL;          /* table of processes */
Pa_Composition *pa_compositionTable = NULL;  /* table of compositions */

/* INTERNAL VARIABLES */

/* used and not initialized by flex/bison */
char *patext = NULL; /* from lex.pa.c */
void *paleng = NULL; /* from lex.pa.c */
int panerrs = 0;     /* from pa.tab.c: number of syntax errors so far */
int pachar = 0;      /* from pa.tab.c: the lookahead symbol */

/* the semantic value of the lookahead symbol */
/* this should be compatible with the union from paParser.y */
struct pastype {
  void *i1;
} palval = {NULL};

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static int wordNumber(Est_String s);

static Est_String shortname(Est_String name);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/* To simplify building on MS WINDOWS - Pa_Init is stated in *.def file */
/* This funcion is defined in tcl part of the package, only */
/* Here, we make it always defined it even if we do not need it */

#ifndef TCL
   void Pa_Init(void) {}
#endif

/**Function****************************************************************
  Synopsis    [Function Pa_InitPkg.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_InitPkg()
{
  if (pa_status == 1) {
    printf("\nProcess algebra package is already initialized. ");
    return;
  } else {
    pa_status = 1;
  }

  pa_variables = 0;
  pa_sorts = 0;
  pa_processes = 0;
  pa_compositions = 0;

  /* INITIALIZATION OF VARIABLE TABLE */
  /* BDD VARIABLE WITH INDEX 0 IS CONSTANT 1 */
  if (!(pa_variableTable = (Pa_Variable *) malloc(sizeof(Pa_Variable)))) {
      fprintf(stderr,"\nPa_InitPkg: Out of memory error!\n");
      exit(1);
  }
  pa_variableTable[0].weight = -1;
  pa_variableTable[0].proccomp = -1;
}

/**Function****************************************************************
  Synopsis    [Function Pa_ExitPkg.]
  Description [Pa_ExitPkg deallocates all memory allocated by pa module.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_ExitPkg()
{
  int i,j;
  Pa_Sort *s;
  Pa_Process *p;
  Pa_Composition *c;

  if (pa_status == 0) {
    printf("\nPA package is not initialized. ");
    return;
  } else {
    pa_status = 0;
  }

  /* DELETE VARIABLES */
  /*
  printf("Delete variables...\n");
  */
  if (pa_variableTable) free(pa_variableTable);

  /* DELETE SORTS */
  /*
  printf("Delete sorts...\n");
  */
  for(i=0;i<pa_sorts;i++) {
    s = &pa_sortTable[i];
    free(s->name);
    if (!s->deleted) {
      for(i=0; i<s->numActions; i++) {
        free(s->table[i].name);
      }
      free(s->table);
    }
  }
  if (pa_sortTable) free(pa_sortTable);
  pa_sorts = 0;

  /* DELETE PROCESSES */
  /*
  printf("Delete processes...\n");
  */
  for(i=0;i<pa_processes;i++) {
    p = &pa_processTable[i];
    free(p->name);
    if (!p->deleted) {
      if (p->compname) free(p->compname);
      for(j=0; j<p->numStates; j++) {
        free(p->tableS[j].name);
        if (p->tableS[j].ccs) {
          if (p->tableS[j].ccs) free(p->tableS[j].ccs);
        }
      }
      if (p->tableS) free(p->tableS);
      if (p->tableT) free(p->tableT);
    }
  }
  if (pa_processTable) free(pa_processTable);
  pa_processes = 0;

  /* DELETE COMPOSITIONS */
  /*
  printf("Delete compositions...\n");
  */
  for(i=0;i<pa_compositions;i++) {
    c = &pa_compositionTable[i];
    free(c->name);
    if (c->tableP) free(c->tableP);
    if (c->tableA) free(c->tableA);
    if (c->di) free(c->di);
    if (c->dij) free(c->dij);
    if (c->alphabet) free(c->alphabet);
  }
  if (pa_compositionTable) free(pa_compositionTable);
  pa_compositions = 0;
}

/**Function****************************************************************
  Synopsis    [Function Pa_AboutPkg]
  Description [Pa_AboutPkg reports release of Process algebra package.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_AboutPkg() {
  printf("EST ");
  printf(EDITION);
  printf(", version ");
  printf(VERSION);
  printf(", ");
  printf("Process algebra package");
  printf(NEWLINE);
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_InitVariable(unsigned short int i)
{

  /* VARIABLE TABLE IS NEVER EMPTY. AT LEAST, THERE IS ELEMENT '1' AT INDEX [0] */
  if (i > pa_variables) {
    pa_variables = i;
    if (!(pa_variableTable = (Pa_Variable *)
        realloc(pa_variableTable, (pa_variables+1) * sizeof(Pa_Variable)))) {
      fprintf(stderr,"\nPa_InitVariable: Out of memory error!\n");
      exit(1);
    }
  }

  pa_variableTable[i].weight = -1;    /*  weight = 1,2,4... (decoding )  */
  pa_variableTable[i].proccomp = -1;   /* process-composition (decoding) */
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_AddNewSort(Est_String name)
{
  Pa_Sort *s;
  int i;
  int sortnum;
  Est_Boolean OK;

  i = 0;
  OK = FALSE;
  if (pa_sorts > 0) {
    while (!OK && i<pa_sorts) {
      if (!strcmp(name,pa_sortTable[i].name)) {
        OK = TRUE;
      } else {
        i++;
      }
    }
  }

  if (pa_sorts == 0) {
    pa_sortTable = (Pa_Sort *) malloc (sizeof(Pa_Sort));
    s = pa_sortTable;
    pa_sorts = 1;
    sortnum = 0;
  } else {

    if (OK) {
      sortnum = i;
      Pa_DeleteSort(i);
      s = &pa_sortTable[i];
      free(s->name); /* deleted sorts have name! */
      /* DEBUGGING */
      /*
      fprintf(stderr,"Pa_AddNewSort: sort %s overwritten\n",name);
      */
    } else {
      sortnum = pa_sorts;
      pa_sorts++;
      pa_sortTable = (Pa_Sort *) realloc (pa_sortTable, pa_sorts * sizeof(Pa_Sort));
      s = &pa_sortTable[pa_sorts-1];
    }
  }

  s->name = strdup(name);
  s->deleted = FALSE;
  s->encoded = 0;
  s->numVar = 0;
  s->numActions = 0;
  s->table = NULL;
  s->a0 = bdd_termNull;

  /* ADD ACTIONS 'TAU' and 'FINAL' */
  /* FINAL is used for WCA */
  s->numActions = 2;
  s->table = (Pa_Action *) malloc (2 * sizeof(Pa_Action));
  s->table[0].name = strdup("TAU");
  s->table[0].p = bdd_termNull;
  s->table[1].name = strdup("FINAL");
  s->table[1].p = bdd_termNull;

  return sortnum;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindSort(Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindSort got null string!\n");
    return -1;
  }
  while (i<pa_sorts) {
    if (!strcmp(name,pa_sortTable[i].name))
      return i;
    else
      i++;
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FOASortAction(Pa_Sort *sort, Est_String name)
{
  Pa_Action *a;
  int i;

  if (!sort) return -1;

  if (!name) {
    printf("\nERROR (Pa_FOASortAction): Null string!\n");
    return -1;
  }

  /* SORT CANNOT BE EMPTY. ALWAYS EXISTS AT LEAST ACTION 'TAU' */

  if (!strcasecmp(name,"TAU")) {
    return 0;
  }

  i = 1;
  while (i<sort->numActions) {
    if (!strcmp(name,sort->table[i].name))
      return i;
    else
      i++;
  }

  if (sort->encoded) {
    printf("\nERROR (Pa_FOASortAction): You cannot add action %s to an encoded sort %s\n",name,sort->name);
    return -1;
  }

  (sort->numActions)++;
  if (!(sort->table = (Pa_Action *) realloc(
                      sort->table, sort->numActions * sizeof(Pa_Action))))
  {
    printf("\nERROR (Pa_FOASortAction): Out of memory!\n");
    return -1;
  }

  a = &sort->table[sort->numActions-1];

  a->name = strdup(name);
  a->p = bdd_termNull;

  return (sort->numActions-1);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindSortAction(Pa_Sort *sort, Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindSortAction got null string!\n");
    return -1;
  }
  if (!sort) return -1;

  while (i<sort->numActions) {
    if (!strcmp(name,sort->table[i].name))
      return i;
    else
      i++;
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_AddNewProcess(Est_String name)
{
  Pa_Process *p;
  int i;
  int procnum;
  Est_Boolean OK;

  i = 0;
  OK = FALSE;
  if (pa_processes > 0) {
    while (!OK && i<pa_processes) {
      if (!strcmp(name,pa_processTable[i].name)) {
        OK = TRUE;
      } else {
        i++;
      }
    }
  }

  if (OK) {

    printf(" (ignored and using the existing instance)");
    procnum = -1;

  } else {
      
    if (pa_processes == 0) {
      procnum = 0;
      pa_processes = 1;
      pa_processTable = (Pa_Process *) malloc (sizeof(Pa_Process));
      p = pa_processTable;
    } else {
      procnum = pa_processes;
      pa_processes++;
      pa_processTable = (Pa_Process *) realloc (
                         pa_processTable, pa_processes * sizeof(Pa_Process));
      p = &pa_processTable[pa_processes-1];
    }

    p->name = strdup(name);
    p->compname = NULL;
    p->encoded = 0;
    p->sort = -1;
    p->numStates = 0;
    p->tableS = NULL;
    p->initial = -1;
    p->stateBDD = bdd_termNull;
    p->numTransitions = 0;
    p->tableT = NULL;
    p->numVar = 0;
    p->d = bdd_termNull;
    p->stab = bdd_termNull;
    p->stablist = bdd_termNull;
    p->trace = FALSE;
    p->weak = FALSE;
    p->strong = FALSE;
    p->deleted = FALSE;
  }

  return procnum;
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindProcess(Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindProcess got null string!\n");
    return -1;
  }

  while (i<pa_processes) {
    if (!strcmp(name,pa_processTable[i].name)) {

      /* DEBUGGING */
      /*
      fprintf(stderr,"Pa_FindProcess: %s (%d)\n",name,i);
      */

      return i;
    } else {
      i++;
    }
  }

  /* DEBUGGING */
  /*
  fprintf(stderr,"Pa_FindProcess: %s (-1)\n",name);
  */

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Renamed process has the same seq as the original one.]
  SeeAlso     []
  ************************************************************************/

void
Pa_RenameProcess(Est_String old, Est_String new)
{
  int oldproc,newproc;
  int i,j;
  Pa_Process *p;
  Est_String state;
  Est_String sup2;

  oldproc = Pa_FindProcess(old);
  if (oldproc == -1) {
    /*
    printf("Process does not exists.\n");
    */
    return;
  }

  newproc = Pa_FindProcess(new);
  if (newproc != -1) {
    /*
    printf(" (process overwritten) ");
    */
    Pa_DeleteProcess(newproc);
  }

  /* DEBUGGING */
  /*
  fprintf(stderr,"Pa_RenameProcess: %s->%s (%d/%d)\n",old,new,oldproc,newproc);
  */

  /* partially composed process does not have transition table, */
  /* but all processes should have state table! */

  if (!pa_processTable[oldproc].tableS) {
    printf("\n    ERROR (Pa_RenameProcess): Missing state table!");
    return;
  }

  /* CHANGE NAME OF PROCESS */
  free(pa_processTable[oldproc].name); /* even deleted processes have name! */
  pa_processTable[oldproc].name = strdup(new);

  /* CHANGE NAME OF STATES */
  p = &pa_processTable[oldproc];
  for (i=0; i<p->numStates; i++) {
    state = strdup(p->tableS[i].name);
    if (strcmp(state,"")) {

      if (Pa_IsGate(state)) {

        if (!strcmp(state,old)) {
          free(p->tableS[i].name);
          p->tableS[i].name = strdup(new);
        }

      } else {

        /* STATES WILL GET THE NAME OF NEW PROCESS */
        sup2 = (Est_String) malloc(65535);
        j=strcspn(state,"<");
        if (j<strlen(state)) {
          state[j]=0;
          sprintf(sup2,"%s<%s>",state,new);
          state[j]='<';
        } else {
          /* this state does not have the name of process */
          printf("\n    WARNING (Pa_RenameProcess): State %s does not have the name of process!",state);
          sprintf(sup2,"%s",state);
        }
        free(p->tableS[i].name);
        p->tableS[i].name = strdup(sup2);
        free(sup2);

      }

      free(state);

    }
  }

  /* RE-ENCODE TO GET CORRECT NAMES FOR BDD VARIABLES */
  if (p->encoded == 1) {

    if ((p->numTransitions == 0) && !Bdd_isEqv(p->d,bdd_termFalse)) {
      fprintf(stderr,"RENAMING PARTIALY COMPOSED PROCESSES IS NOT IMPLEMENTED, YET\n");
    }

    p->encoded = 0;
    Pa_EncodeProcess(p->name);

  }

  /* DEBUGGING */
  /*
  fprintf(stderr,"RENAME PROCESS FINISH %s\n",p->name);
  */

}

/**Function****************************************************************
  Synopsis    []
  Description [name = original process, copy = new process,
               copysort = sort of new process,
               old = the list of actions to rename,
               new = new names for actions in old list]
  SideEffects [COPY OF NOT-ENCODED PROCESS IS NOT ENCODED!
               COPY OF FULL COMPOSED ENCODED PROCESS IS NOT ENCODED!
               INDEED, COPY OF PARTIALLY COMPOSED PROCESS IS ENCODED,
               BUT IT USES OLD SORT!
               COPYSORT IS NEVER ENCODED!]
  SeeAlso     []
  ************************************************************************/

int
Pa_CopyProcess(Est_String name, Est_String copy, Est_String copysort,
               Est_String old, Est_String new)
{
  int procnum,copynum,oldsort,newsort;
  int n;
  Pa_Process *p,*cp;
  int i,j;
  Est_String state;
  Est_String sup1, sup2, sup1x, sup2x, act;
  int token;
  int *renamingTable;

  procnum = Pa_FindProcess(name);
  if (procnum == -1) {
    /*
    printf("Process does not exists.\n");
    */
    return -1;
  }
  p = &pa_processTable[procnum];

  if ((p->numTransitions == 0) && Bdd_isEqv(p->d,bdd_termFalse)) {
    printf("(empty process)");
  }

  /* DEBUGGING */
  /*
  printf("Pa_CopyProcess: %s (%d) -> %s\n",name,procnum,copy);
  */

  /* partially composed process does not have transition table, */
  /* but it should have state table! */

  if ((p->encoded) && (!p->tableS)) {
    printf("\n    ERROR (Pa_CopyProcess): Missing state table!");
    return -1;
  }

  if (old) {
    sup1x = strdup(old);
  } else {
    sup1x = strdup("");
  }

  if (new) {
    sup2x = strdup(new);
  } else {
    sup2x = strdup("");
  }

  /* PREPARE NEW SORT */
  /* new process will use new sort, it must be an existed one */
  /* it could be empty or non-empty */
  /* if it is encoded then changes (new actions) are not allowed */

  oldsort = p->sort;

  if (!copysort) {

    /* DEBUGGING */
    /*
    if ((p->encoded == 1) && (p->numTransitions == 0) && !Bdd_isEqv(p->d,bdd_termFalse)) {
      printf("\nNOTE: Copying partially composed process %s -> %s without changing sort\n",name,copy);
    } else {
      printf("\nNOTE: Copying process %s -> %s without changing sort\n",name,copy);
    }
    */

    newsort = oldsort;

  } else {

    /* DEBUGGING */
    /*
    if ((p->encoded == 1) && (p->numTransitions == 0) && !Bdd_isEqv(p->d,bdd_termFalse)) {
      printf("\nNOTE: Copying partially composed process %s -> %s using new sort %s\n",name,copy,copysort);
    } else {
      printf("\nNOTE: Copying process %s -> %s using new sort %s\n",name,copy,copysort);
    }
    */

    if ((newsort = Pa_FindSort(copysort)) == -1) {
      printf("\n    ERROR (Pa_CopyProcess): Sort %s does not exists!",copysort);
      return -1;
    }

    /* DEBUGGING */
    /*
    if (pa_sortTable[newsort].encoded) {
      printf("\nNOTE: New sort %s is already encoded",pa_sortTable[newsort].name);
    }
    */

    /* DEBUGGING */
    /*
    if (newsort == oldsort) {
      printf("\nNOTE: New sort is an extension of old sort\n");
    } else {
      printf("\nNOTE: New sort %s is NOT extension of old sort %s\n",pa_sortTable[newsort].name,pa_sortTable[oldsort].name);
    }
    */

  }

  /* ADD NEW PROCESS */

  copynum = Pa_AddNewProcess(copy); /* add new not-encoded process */
  if (copynum == -1) return -1;

  p = &pa_processTable[procnum]; /* AddNewProcess calls realloc(pa_processTable) !! */
  cp = &pa_processTable[copynum];

  if (p->compname) cp->compname = strdup(p->compname);
  cp->sort = newsort;
  cp->initial = p->initial;
  cp->numStates = p->numStates;
  cp->numTransitions = p->numTransitions;
  cp->trace = p->trace;
  cp->weak = p->weak;
  cp->strong = p->strong;

  /* COPY STATES */

  if (p->numStates > 0) {
    cp->tableS = (Pa_State *) malloc(p->numStates * sizeof(Pa_State));

    for (i=0; i<p->numStates; i++) {
      state = strdup(p->tableS[i].name);
      if (!strcmp(state,"")) {

         /*state is deleted */
         cp->tableS[i].name = strdup("");

      } else {

        if (Pa_IsGate(state)) {

          if (!strcmp(state,name)) {
            cp->tableS[i].name = strdup(cp->name);
          } else {
            cp->tableS[i].name = strdup(state);
          }

        } else {

          /* STATES WILL GET THE NAME SUFFIX OF NEW PROCESS */
          sup2 = (Est_String) malloc(65535);
          j=strcspn(state,"<");
          if (j<strlen(state)) {
            state[j]=0;
            sprintf(sup2,"%s<%s>",state,shortname(cp->name)); /* shortname! why? */
            state[j]='<';
          } else {
            /* this state does not have the name of process */
            printf("\n    WARNING (Pa_CopyProcess): State %s does not have the name of process!",state);
            sprintf(sup2,"%s",state);
          }
          cp->tableS[i].name = strdup(sup2);
          free(sup2);

        }
      }

      free(state);

      cp->tableS[i].bddR = bdd_termNull;
      cp->tableS[i].bddS = bdd_termNull;
      cp->tableS[i].bddComp = p->tableS[i].bddComp;
      cp->tableS[i].acc = bdd_termNull;
      cp->tableS[i].minacc = bdd_termNull;
      cp->tableS[i].final = p->tableS[i].final;
      cp->tableS[i].ccs = NULL;
    }
  } else {
    cp->tableS = NULL;
  }

  /* COPY TRANSITIONS - FULL COMPOSED PROCESSES, ONLY */
  /* IN ANY CASE, UPDATE NEW SORT WITH ACTIONS FROM OLD SORT */
  /* partially composed processes have p->numTransitions == 0 */

  if (p->numTransitions > 0) {

    cp->tableT = (Pa_Transition *)
                    malloc (p->numTransitions *sizeof(Pa_Transition));

    for (i=0; i<p->numTransitions; i++) {
      cp->tableT[i].source = p->tableT[i].source;

      if (p->tableT[i].action == 0) {
        /* OLD = TAU */
        cp->tableT[i].action = 0;
      } else {
        if ((cp->tableT[i].action = Pa_FOASortAction(&pa_sortTable[newsort],
             pa_sortTable[oldsort].table[p->tableT[i].action].name)) == -1)
        {
          printf("\n    ERROR (copy %s -> %s): Cannot add action to sort %s (encoded sort?!)",name,copy,copysort);
          free(cp->tableT);
          cp->numTransitions = 0;
          cp->tableT = NULL;
          return -1;
        }
      }

      cp->tableT[i].type = p->tableT[i].type;
      cp->tableT[i].dest = p->tableT[i].dest;
    }

  } else {

    cp->tableT = NULL;

    /* UPDATE NEW SORT - ADD ALL ACTIONS FROM OLD SORT */
    for (i=0; i<pa_sortTable[oldsort].numActions; i++) {
      if ((Pa_FOASortAction(&pa_sortTable[newsort],
           pa_sortTable[oldsort].table[i].name)) == -1)
      {
        printf("\n    ERROR (copy %s -> %s): Cannot add actions to the encoded sort %s",name,copy,copysort);
        return -1;
      }
    }

  }

  /* CHANGING TRANISTIONS BY RENAMING ACTIONS - FULL COMPOSED PROCESSES, ONLY */

  n = wordNumber(sup1x);
  if (n != wordNumber(sup2x)) {
    printf("\n    ERROR: Different number of old and new action names!");
    free(cp->tableT);
    cp->tableT = NULL;
    cp->numTransitions = 0;
    return -1;
  }
  if ((n>0) && (p->numTransitions == 0) && !Bdd_isEqv(p->d,bdd_termFalse)) {
    printf("\n    ERROR: Changing action names in partially composed process is not supported!");
    free(cp->tableT);
    cp->tableT = NULL;
    cp->numTransitions = 0;
    return -1;
  }

  /* IF PROCESS IS NOT ENCODED (ENCODED => EXPANDED) */
  /* THEN ALL GATES MUST BE EXPANDED BEFORE RENAMING */

  if ((n>0) && !p->encoded) {
    Pa_ExpandGates(cp);
  }

  if (n>0) {
    sup1 = sup1x;
    sup2 = sup2x;
    renamingTable = (int *) calloc (2*n,sizeof(int));
  } else {
    renamingTable = NULL;
  }

  for (i=1; i<=n; i++) {
    token = strspn(sup1," ");
    sup1 = &sup1[token];
    token = strcspn(sup1," ");
    act = sup1;
    act[token] = 0;

    /* if (!strcmp(act,"tau") || !strcmp(act,"Tau") || !strcmp(act,"TAU")) { */
    if (!strcasecmp(act,"TAU")) {
      printf("\n    WARNING: Action %s cannot be renamed\n",act);
      renamingTable[2*i-2] = 0;
    } else {
      /* ADD ACTION IF IT DOES NOT EXISTS! */
      if ((renamingTable[2*i-2] = Pa_FOASortAction(&pa_sortTable[newsort],act)) == -1)
      {
        printf("\n    ERROR (rename1): Cannot change/add action %s (encoded sort?!)",act);
        free(sup1x);
        free(sup2x);
        free(renamingTable);
        free(cp->tableT);
        cp->tableT = NULL;
        cp->numTransitions = 0;
        return -1;
      }
    }

    /*
    fprintf(stderr,"Action %s ",act);
    */

    sup1 = &sup1[token+1];

    token = strspn(sup2," ");
    sup2 = &sup2[token];
    token = strcspn(sup2," ");
    act = sup2;
    act[token] = 0;

    /* if (!strcmp(act,"tau") || !strcmp(act,"Tau") || !strcmp(act,"TAU")) { */
    if (!strcasecmp(act,"TAU")) {
      renamingTable[2*i-1] = 0;
    } else {
      /* ADD ACTION IF IT DOES NOT EXISTS! */
      if ((renamingTable[2*i-1] = Pa_FOASortAction(&pa_sortTable[newsort],act)) == -1)
      {
        printf("\n    ERROR (rename2): Cannot change/add action %s (encoded sort?!)",act);
        free(sup1x);
        free(sup2x);
        free(renamingTable);
        free(cp->tableT);
        cp->tableT = NULL;
        cp->numTransitions = 0;
        return -1;
      }
    }

    /*
    fprintf(stderr,"will be renamed to action %s\n",act);
    */

    sup2 = &sup2[token+1];
  }

  free(sup1x);
  free(sup2x);

  /*
  fprintf(stderr,"Pa_CopyProcess\n");
  for (i=0; i<n; i++) {
    fprintf(stderr,"Action number %d will be ",renamingTable[2*i]);
    fprintf(stderr,"renamed to action number %d.\n",renamingTable[2*i+1]);
  }
  */

  /* COPY TRANSITION RELATION */

  if ((p->encoded == 1) && (p->numTransitions == 0) && !Bdd_isEqv(p->d,bdd_termFalse)) {

    Bdd_Edge substPR,substQS,sup1,sup2;

    /* DEBUGGING */
    /*
    printf("\nNOTE: Processing partially composed process.");
    */

    /* encode states, only */
    /* keep transition relation */
    /* sort must not be encoded! */

    PaEncodeProcessStates(cp);

    /* transition relation is changed to use new process variables */

    substPR = bdd_termFalse;
    for (i=0; i<p->numStates; i++) {
      sup1 = Bdd_RelOpSimple(p->tableS[i].bddR,"R2P",TRUE);
      sup2 = Bdd_ITE(cp->tableS[i].bddR,sup1,bdd_termFalse);
      substPR = Bdd_ITE(substPR,bdd_termTrue,sup2);
    }

    sup1 = Bdd_RelOpSimple(p->d,"R2P S2Q",TRUE);
    sup2 = Bdd_RelOp(substPR,sup1,"#AND Ex xP",TRUE);

    substQS = Bdd_RelOpSimple(substPR,"P2Q R2S",TRUE);
    cp->d = Bdd_RelOp(substQS,sup2,"#AND Ex xQ",TRUE);

    cp->encoded = 1;
    Bdd_SaveFormula("D_",copy,cp->d);

    /* COPYSORT MUST NOT BE ENCODED */
    /* THUS, TRANSITION RELATION STILL DEPENDS ON OLD SORT !!! */
    cp->sort = oldsort;

  } else {

    /* DEBUGGING */
    /*
    if (pa_processTable[procnum].encoded) {
      printf("\nNOTE: Processing encoded full composed process.");
    } else {
      printf("\nNOTE: Processing non-encoded full composed process.");
    }
    */

    /* RENAME ACTIONS IN TRANSITION TABLE */
    /* THIS IS THE SAME FOR NOT-ENCODED AND ENCODED FULL PROCESS */

    if (n>0) {
      for (i=0; i<cp->numTransitions; i++) {
        if (cp->tableT[i].action != 0) {
          for (j=0; j<n; j++) {
            if (renamingTable[2*j] == cp->tableT[i].action) {
              /* ACTION WILL BE RENAMED */
              cp->tableT[i].action = renamingTable[2*j+1];
              /* NEW = TAU */
              if (cp->tableT[i].action == 0) {
                cp->tableT[i].type = 0; /* THIS IS TYPE OF ACTION TAU */
              }
            }
          }
        }
      }
    }

  }

  if (renamingTable) free(renamingTable);

  return copynum;
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Difference between state and gate is the name!
               Gates are closely related to implementation of CCS parser.]
  SeeAlso     []
  ************************************************************************/

Est_Boolean
Pa_IsGate(Est_String name)
{
  Est_Boolean r;

  r = FALSE;

  if (!name) {
    printf("\nError: Pa_IsGate got null string!\n");
    return r;
  }

  if (!strcmp(name,"")) {
    r = FALSE;
  } else {
    if (strchr(name,'<')) {
      r = FALSE;
    } else {
      r = TRUE;
    }
  }

  return r;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Suffix '<'+proc->name+'>' is added]
  SeeAlso     []
  ************************************************************************/

int
Pa_FOAStateProcess(Pa_Process *proc, Est_String name)
{
  Est_String name1;
  Pa_State *s;
  int i;

  if (!proc) {
    printf("\nError: Pa_FOAStateProcess got null proc!\n");
    return -1;
  }

  if (!name) {
    printf("\nError: Pa_FOAStateProcess got null name!\n");
    return -1;
  }

  if (!strcmp(name,"")) {
     name1 = strdup(name);
  } else {
    if (strchr(name,'<')) {
      name1 = strdup(name);
    } else {
      name1 = (Est_String) malloc(127);
      sprintf(name1,"%s<%s>",name,shortname(proc->name)); /* shortname! why? */
    }
  }

  i = 0;
  if (proc->numStates == 0) {
    proc->numStates = 1;
    proc->tableS = (Pa_State *) malloc (sizeof(Pa_State));
    s = proc->tableS;
  } else {
    while (i<proc->numStates) {
      if (!strcmp(name1,proc->tableS[i].name)) {
        free (name1);
        return i;
      } else {
        i++;
      }
    }

    if (proc->encoded) {
      printf("\nError: You cannot add states to an encoded process!\n");
      free (name1);
      return -1;
    }

    (proc->numStates)++;
    proc->tableS = (Pa_State *) realloc (proc->tableS,
                                    proc->numStates * sizeof(Pa_State));
    s = &proc->tableS[proc->numStates-1];
  }

  s->name = strdup(name1);
  s->bddR = bdd_termNull;
  s->bddS = bdd_termNull;
  s->bddComp = bdd_termNull;
  s->acc = bdd_termNull;
  s->minacc = bdd_termNull;
  s->final = FALSE;
  s->ccs = NULL;

  free (name1);
  return (proc->numStates-1);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Almost copy of FOAState.]
  SeeAlso     []
  ************************************************************************/

int
Pa_FOAGateProcess(Pa_Process *proc, Est_String name)
{
  Est_String name1;
  Pa_State *s;
  int i;

  if (!proc) {
    printf("\nError: Pa_FOAGateProcess got null proc!\n");
    return -1;
  }

  if (!name) {
    printf("\nError: Pa_FOAGateProcess got null string!\n");
    return -1;
  }

  if (strchr(name,'<')) {
    printf("\nError: The name of GATE must not contain '<>'!\n");
    return -1;
  } else {
    name1 = strdup(name);
  }

  i = 0;
  if (proc->numStates == 0) {
    proc->numStates = 1;
    proc->tableS = (Pa_State *) malloc (sizeof(Pa_State));
    s = proc->tableS;
  } else {
    while (i<proc->numStates) {
      if (!strcmp(name1,proc->tableS[i].name)) {
        free (name1);
        return i;
      } else {
        i++;
      }
    }

    if (proc->encoded) {
      printf("\nError: You cannot add gates to an encoded process!\n");
      free (name1);
      return -1;
    }

    (proc->numStates)++;
    proc->tableS = (Pa_State *) realloc (proc->tableS,
                                   proc->numStates * sizeof(Pa_State));
    s = &proc->tableS[proc->numStates-1];
  }

  s->name = strdup(name1);
  s->bddR = bdd_termNull;
  s->bddS = bdd_termNull;
  s->bddComp = bdd_termNull;
  s->acc = bdd_termNull;
  s->minacc = bdd_termNull;
  s->final = FALSE;
  s->ccs = NULL;

  free (name1);
  return (proc->numStates-1);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FOAComposedStateProcess(Pa_Process *proc, Bdd_Edge F, Est_String prefix)
{
  int i = 0;

  if (Bdd_isNull(F)) {
    printf("\nError: Pa_FOAComposedStateProcess got bdd_termNull!\n");
    return -1;
  }

  if (Bdd_isEqv(F,bdd_termFalse)) {
    printf("\nError: Pa_FOAComposedStateProcess got bdd_termFalse!\n");
    return -1;
  }

  if (Bdd_isEqv(F,bdd_termTrue)) {
    printf("\nError: Pa_FOAComposedStateProcess got bdd_termTrue!\n");
    return -1;
  }

  if (!proc) return -1;
  if (proc->numStates > 0) {
    while (i<proc->numStates) {
      if (Bdd_isEqv(F,proc->tableS[i].bddComp)) {
        return i;
      } else {
        i++;
      }
    }
  }

  /* if !prefix then add is not allowed */
  if (!prefix) {
    i = -1;
  } else {
    i = Pa_FOANextStateProcess(proc,prefix);
    proc->tableS[i].bddComp = F;
  }

  return i;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindStateProcess(Pa_Process *proc, Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindStateProcess got null string!\n");
    return -1;
  }

  if (!proc) return -1;

  i = 0;
  while (i<proc->numStates) {
    if (!strcmp(name,proc->tableS[i].name)) {
      return i;
    } else {
      i++;
    }
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Suffix after '<' is removed, '<'+proc->name+'>' is added]
  SeeAlso     []
  ************************************************************************/

int
Pa_AddStateProcess(Pa_Process *proc, Est_String name)
{
  Est_String name1,name2;
  int i = 0;

  if (!name) {
    printf("\nError: Pa_AddStateProcess got null string!\n");
    return -1;
  }

  if (strchr(name,'<')) {
    i = strcspn(name,"<");
    name1 = (Est_String) malloc(i+1);
    memmove(name1,name,i);
    name1[i]=0;
  } else {
    name1 = strdup(name);
  }

  name2 = (Est_String) malloc(255);
  sprintf(name2,"%s<%s>",name1,shortname(proc->name)); /*ADDED shortname */
  while (Pa_FindStateProcess(proc,name2) != -1) {
    i = strlen(name1);
    name1 = (Est_String) realloc(name1,i+2);
    name1[i] = 0x27;
    name1[i+1] = 0x00;
    sprintf(name2,"%s<%s>",name1,shortname(proc->name)); /*ADDED shortname */
  }

  i = Pa_FOAStateProcess(proc,name2);

  free(name1);
  free(name2);
  return i;
}

/**Function****************************************************************
  Synopsis    []
  Description [Add a state to the proces. The name of the state will be
               composed of given prefix and a number.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FOANextStateProcess(Pa_Process *proc, Est_String prefix)
{
  int i;
  Est_String x;

  if (!prefix) {
    printf("\nError: Pa_FOANextStateProcess got null string!\n");
    return -1;
  }

  x = (Est_String) malloc(127);
  sprintf(x,"%s%d<%s>",prefix,proc->numStates+1,shortname(proc->name)); /*ADDED shortname */

  i = Pa_FOAStateProcess(proc,x);

  free(x);
  return i;
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_NumberStateProcess(Pa_Process *proc)
{
  int i;
  int n;

  n = 0;
  for(i=0; i<proc->numStates; i++) {
    if (strcmp(proc->tableS[i].name,"")) {
      if (!Pa_IsGate(proc->tableS[i].name)) {
        n++;
      }
    }
  }

  return n;
}

/**Function****************************************************************
  Synopsis    []
  Description [type 0 = output, type 1 = input]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FOATransition(Pa_Process *proc, int state1, int action, int type,
                 int state2)
{
  Pa_Transition *t;
  int i = 0;

  if (!proc) return -1;
  if (proc->numTransitions == 0) {
    proc->numTransitions = 1;
    proc->tableT = (Pa_Transition *) malloc (sizeof(Pa_Transition));
    t = proc->tableT;
  } else {
    while (i<proc->numTransitions) {
      if ((state1 == proc->tableT[i].source) &&
          (action == proc->tableT[i].action) &&
          (type == proc->tableT[i].type) &&
          (state2 == proc->tableT[i].dest))
      {
        return i;
      } else {
        i++;
      }
    }

    if (proc->encoded) {
      printf("\nError: You cannot add transitions to an encoded process!\n");
      return -1;
    }

    (proc->numTransitions)++;
    proc->tableT = (Pa_Transition *) realloc (proc->tableT,
                                proc->numTransitions * sizeof(Pa_Transition));

    t = &proc->tableT[proc->numTransitions-1];
  }

  t->source = state1;
  t->action = action;
  t->type = type;
  t->dest = state2;

  return (proc->numTransitions-1);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_AddNewComposition(Est_String name, int type, int numP, int numA)
{
  Pa_Composition *c;
  int i;
  int compnum;
  Est_Boolean OK;

  i = 0;
  OK = FALSE;
  if (pa_compositions > 0) {
    while (!OK && i<pa_compositions) {
      if (!strcmp(name,pa_compositionTable[i].name)) {
        OK = TRUE;
      } else {
        i++;
      }
    }
  }

  if (OK) {

    printf(" (ignored and using the existing instance)");
    compnum = -1;

  } else {

    if (pa_compositions == 0) {
      compnum = 0;
      pa_compositions = 1;
      pa_compositionTable = (Pa_Composition *) malloc (sizeof(Pa_Composition));
      c = pa_compositionTable;
    } else {
      compnum = pa_compositions;
      pa_compositions++;
      pa_compositionTable = (Pa_Composition *) realloc (pa_compositionTable,
                             pa_compositions * sizeof(Pa_Composition));
      c = &pa_compositionTable[compnum];
    }

    c->name = strdup(name);
    c->type = type;
    c->numProcesses = numP;
    c->tableP = (int *) calloc (numP,sizeof(int));
    c->numActions = numA;
    if (numA > 0) {
      c->tableA = (int *) calloc (numA,sizeof(int));
    } else {
      c->tableA = NULL;
    }
    c->sort = -1;
    c->initialBDD = bdd_termNull;
    c->stateBDD = bdd_termNull;
    c->deadlockedBDD = bdd_termNull;
    c->transitionBDD = bdd_termNull;
    c->actionsBDD = bdd_termNull;
    c->di = NULL;
    c->dij = NULL;
    c->alphabet = NULL;
    c->snew = bdd_termNull;
    c->dinew = NULL;
    c->dijnew = NULL;
    c->dicube = NULL;
    c->dijcube = NULL;
    c->onthefly = FALSE;
    c->full = FALSE;
    c->trace = FALSE;
    c->weak = FALSE;
    c->strong = FALSE;
  }

  return (compnum);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [The state is only erased and not deleted from the table.
               Positions of states in the state table are not changed.]
  SeeAlso     []
  ************************************************************************/

void
Pa_DeleteStateProcess(Pa_Process *proc, int num)
{
  Pa_State *s;

  s = &proc->tableS[num];

  if (s->name) {
    free(s->name);
    s->name = strdup("");
  }
  s->bddR = bdd_termNull;
  s->bddS = bdd_termNull;
  s->bddComp = bdd_termNull;
  s->acc = bdd_termNull;
  s->minacc = bdd_termNull;
  s->final = FALSE;
  if (s->ccs) {
    free(s->ccs);
    s->ccs = NULL;
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Sort is erased and not removed from the table!
               The sequence numbers of other sorts are not changed.]
  SeeAlso     []
  ************************************************************************/

void
Pa_DeleteSort(int num)
{
  int i;
  Pa_Sort *s;

  if (num == -1) return;
  s = &pa_sortTable[num];

  /* REMOVE FROM BDD FORMULA TABLE */
  for(i=0; i<s->numActions; i++) {
    s->table[i].p = bdd_termNull;
  }
  Pa_SaveSortActions(s);

  /* CLEAN */
  for(i=0; i<s->numActions; i++) {
    if (s->table[i].name) {
      free(s->table[i].name);
      s->table[i].name = NULL;
    }
  }
  free(s->table);

  s->deleted = TRUE;
  s->encoded = 0;
  s->numVar = 0;
  s->numActions = 0;
  s->table = NULL;
  s->a0 = bdd_termNull;

  if (s->name) free(s->name);
  s->name = strdup("__DELETED__");
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [The sequence numbers of sorts can be changed!]
  SeeAlso     []
  ************************************************************************/

void
Pa_RemoveDeletedSorts()
{
  int i;

  i = 0;
  while (i < pa_sorts) {
    if (pa_sortTable[i].deleted) {
      free(pa_sortTable[i].name);
      if (pa_sorts == 1) {
        free(pa_sortTable);
        pa_sorts = 0;
      } else {
        pa_sorts--;
        if (i != pa_sorts) {
          pa_sortTable[i] = pa_sortTable[pa_sorts];
        }
        pa_sortTable = (Pa_Sort *) realloc (pa_sortTable,
                        pa_sorts * sizeof(Pa_Sort));
      }
    } else {
      i++;
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [Process is erased and not removed from the table!
               The sequence numbers of other processes are not changed.]
  SeeAlso     []
  ************************************************************************/

void
Pa_DeleteProcess(int num)
{
  int i;
  Pa_Process *p;

  if (num == -1) return;
  p = &pa_processTable[num];

  /* CLEAN FORMULAE TREE */
  p->d = bdd_termNull;
  p->stateBDD = bdd_termNull;
  p->stab = bdd_termNull;
  p->stablist = bdd_termNull;
  Bdd_SaveFormula("D_",p->name,bdd_termNull);
  Bdd_SaveFormula("Initial_",p->name,bdd_termNull);
  Bdd_SaveFormula("S_",p->name,bdd_termNull);
  Bdd_SaveFormula("STAB_",p->name,bdd_termNull);
  Bdd_SaveFormula("STABLIST_",p->name,bdd_termNull);
  for(i=0; i<p->numStates; i++) {
    p->tableS[i].bddR = bdd_termNull;
    p->tableS[i].bddS = bdd_termNull;
    p->tableS[i].bddComp = bdd_termNull;
  }
  Pa_SaveProcStates(p);

  /* CLEAN */
  for(i=0; i<p->numStates; i++) {
    if (p->tableS[i].name) {
      free(p->tableS[i].name);
      p->tableS[i].name = NULL;
    }
    if (p->tableS[i].ccs) {
      free(p->tableS[i].ccs);
      p->tableS[i].ccs = NULL;
    }
  }
  if (p->tableS) free(p->tableS);
  if (p->tableT) free(p->tableT);

  p->deleted = TRUE;
  p->encoded = 0;
  p->sort = -1;
  p->numStates = 0;
  p->tableS = NULL;
  p->initial = -1;
  p->numTransitions = 0;
  p->tableT = NULL;
  p->numVar = 0;

  if (p->name) free(p->name);
  p->name = strdup("__DELETED__");
  if (p->compname) free(p->compname);
  p->compname = NULL;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [The sequence numbers of processes can be changed!]
  SeeAlso     []
  ************************************************************************/

void
Pa_RemoveDeletedProcesses()
{
  int i;

  i = 0;
  while (i < pa_processes) {
    if (pa_processTable[i].deleted) {
      free(pa_processTable[i].name);
      if (pa_processes == 1) {
        free(pa_processTable);
        pa_processTable = NULL;
        pa_processes = 0;
      } else {
        pa_processes--;
        if (i != pa_processes) {
          pa_processTable[i] = pa_processTable[pa_processes];
        }
        pa_processTable = (Pa_Process *) realloc (pa_processTable,
                           pa_processes * sizeof(Pa_Process));
      }
    } else {
      i++;
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_RemoveComposition(int num)
{
  Pa_Composition *c;

  if (num == -1) return;
  c = &pa_compositionTable[num];

  /* CLEAN FORMULAE TREE */
  Bdd_SaveFormula("Initial_COMP_",c->name,bdd_termNull);
  Bdd_SaveFormula("S_COMP_",c->name,bdd_termNull);
  Bdd_SaveFormula("SD_COMP_",c->name,bdd_termNull);
  Bdd_SaveFormula("D_COMP_",c->name,bdd_termNull);

  /* CLEAN */
  if (c->tableP) free(c->tableP);
  if (c->tableA) free(c->tableA);
  if (c->di) free(c->di);
  if (c->dij) free(c->dij);
  if (c->dinew) free(c->dinew);
  if (c->dijnew) free(c->dijnew);
  if (c->dicube) free(c->dicube);
  if (c->dijcube) free(c->dijcube);
  if (c->alphabet) free(c->alphabet);
  if (c->name) free(c->name);

  /* DELETE */
  if (pa_compositions == 1) {
    free(pa_compositionTable);
    pa_compositionTable = NULL;
    pa_compositions = 0;
  } else {
    pa_compositions--;
    if (num != pa_compositions) {
      pa_compositionTable[num] = pa_compositionTable[pa_compositions];
    }
    pa_compositionTable = (Pa_Composition *) realloc (pa_compositionTable,
                            pa_compositions * sizeof(Pa_Composition));
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindComposition(Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindComposition got null string!\n");
    return -1;
  }

  while (i<pa_compositions) {
    if (!strcmp(name,pa_compositionTable[i].name)) {
      return i;
    } else {
      i++;
    }
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_FindProcessComp(Pa_Composition *c, Est_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Pa_FindProcessComp got null string!\n");
    return -1;
  }

  if (!c) return -1;

  while (i<c->numProcesses) {
    if (!strcmp(name,pa_processTable[c->tableP[i]].name))
      return i;
    else
      i++;
  }

  return -1;
}

/**Function****************************************************************
  Synopsis    []
  Description [Add stabs with bdd-based manipulation.]
  SideEffects [THIS FUNCTION IS NOT IMPLEMENTED - DO NOT USE IT!]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Pa_AddStab(Bdd_Edge d, Est_String name)
{
  /* ADD stabs WITH BDD-BASED MANIPULATION */
  /* THIS FUNCTION SHOULD USE stablist, SEE SIMILAR FUNCTION IN bdd MODUL */
  /* MAYBE, IT CAN DIRECTLY USE SIMILAR FUNCTION IN bdd MODUL */

  fprintf(stderr,"FUNCTION Pa_AddStab IS NOT IMPLEMENTED - DO NOT USE IT!\n");
  exit(1);

  return bdd_termNull;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_SaveSortActions(Pa_Sort *s)
{
  int i;
  Est_String x;

  /* ACTION TAU */
  x = (Est_String) malloc(5 + strlen(s->name));
  sprintf(x,"%s_TAU",s->name);
  Bdd_SaveFormula("Action_",x,s->table[0].p);
  free(x);

  /* OTHER ACTIONS */
  for (i=1; i<s->numActions; i++) {

    /* SAVE ACTIONS WITHOUT TYPE */
    x = (Est_String) malloc(2 + strlen(s->name) + strlen(s->table[i].name));
    sprintf(x,"%s_%s",s->name,s->table[i].name);
    Bdd_SaveFormula("Action_",x,s->table[i].p);

    /* SAVE ACTIONS WITH TYPE - NOT NEEDED */
    /*
    x = (Est_String) malloc(3 + strlen(s->name) + strlen(s->table[i].name));
    sprintf(x,"%s_%s!",s->name,s->table[i].name);
    Bdd_SaveFormula("Action_",x,Bdd_ITE(s->a0,s->table[i].p,bdd_termFalse));
    x[strlen(x)-1] = '?';
    Bdd_SaveFormula("Action_",x,Bdd_ITE(s->a0,bdd_termFalse,s->table[i].p));
    free(x);
    */
  }

}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_SaveProcStates(Pa_Process *p)
{
  int i;
  Est_String x;

  for (i=0; i<(p->numStates); i++) {
    x = (Est_String) malloc(2 + strlen(p->name) + strlen(p->tableS[i].name));
    sprintf(x,"%s_%s",p->name,p->tableS[i].name);
    Bdd_SaveFormula("State_",x,p->tableS[i].bddR);
    Bdd_SaveFormula("NextState_",x,p->tableS[i].bddS);
    Bdd_SaveFormula("CompState_",x,p->tableS[i].bddComp);
    free(x);
  }
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
wordNumber(Est_String s)
{
  int num;
  int token;
  Est_String sup,sup_next;

  num = 0;
  sup = strdup(s);

  token = strspn(sup," ");
  while (strspn(sup," ") < strlen(sup)) {
    sup_next = (Est_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    num++;
    token = strcspn(sup," ");
    sup_next = (Est_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    token = strspn(sup," ");
  }

  free(sup);
  return num;
}

/**Function****************************************************************
  Synopsis    [Function shortname]
  Description [Remove text after '#']
  SideEffects [This may be different than in other files.]
  SeeAlso     []
  ************************************************************************/

static Est_String
shortname(Est_String name) {
  int i;
  Est_String shortname;

  i = strcspn(name,"#");
  shortname = (char *) malloc(i+1);
  memmove(shortname,name,i);
  shortname[i]=0;
  return shortname;
}
