/**CFile*******************************************************************
  PackageName [versis]
  Synopsis    [Package 'versis' provides basic tools for formal
               verification of concurrent systems]

  FileName    [versisComp.c]
  Revision    [$Revision: 126 $]
  Date        [$Date: 2016-07-12 20:51:30 +0200 (tor, 12 jul 2016) $]
  Authors     [Robert Meolic (robert.meolic@um.si),
               Mirjam Sepesy Maucec (mirjam.sepesy@um.si),
               Tatjana Kapus (tatjana.kapus@um.si)]
  Description [File versisComp.c provides parallel composition.]
  SeeAlso     [versis.h, versisInt.h]

  Copyright   [This file is part of EST (Efficient Symbolic Tools).
               Copyright (C) 2003, 2016
               UM-FERI, Smetanova ulica 17, SI-2000 Maribor, Slovenia

               EST is free software; you can redistribute it and/or modify
               it under the terms of the GNU General Public License as
               published by the Free Software Foundation; either version 2
               of the License, or (at your option) any later version.

               EST is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
               Boston, MA 02110-1301 USA.]
  ************************************************************************/

#include "versisInt.h"

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static int wordNumber(Est_String s);

static Est_String shortname(Est_String name);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Versis_Compose]
  Description [0 - CCS compose general (using restricted actions),
               1 - CCS compose with non-restricted actions,
               2 - CCS compose with restricted actions,
               3 - multi-way general (using restricted actions),
               4 - multi-way with non-restricted actions,
               5 - multi-way with restricted actions,
               6 - fundamental-mode multi-way general (using restr. act.),
               7 - fundamental-mode multi-way with non-restricted actions,
               8 - fundamental-mode multi-way with restricted actions

               general (0) = given actions sync others sync and non-sync
               non-restricted actions (1) = they and tau non-sync, others sync
               restricted actions (2) = they sync, others non-sync
               multi-way = more than two processes can synchronise
               fundamental-mode = delete inputs if there is an output

               CCS COMPOSITION: type = 0, actions = given
               LOTOS SYNCHRONISING (only tau goes alone): type = 1, actions = NULL
               LOTOS INTERLEAVING (all actions go alone, only): type = 2, actions = NULL
              ]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Versis_Compose(Est_String name, int type, Est_String process,
               Est_String actions, Est_Boolean onthefly)
{
  int i,j,k,n,m;
  int numP,numA,numcomp;
  int token;
  Pa_Composition *c;
  Pa_Process *p,*cp;
  Est_String sup,sup_next;
  Est_String s;
  Bdd_Edge tau;
  Bdd_Edge sup1,sup2,supinout;
  Est_Boolean minimize;

  /* DEBUGGING */
  /*
  printf("Compose: <%s>\n",name);
  printf("Processes: <%s>\n",process);
  printf("Actions: <%s>\n",actions);
  */

  numP = wordNumber(process);
  numA = wordNumber(actions);

  if ( numP < 1) {
    printf(" ERROR: there are no processes to compose.\n");
    free((void *)1); /* BREAKPOINT */
    return;
  }

  if ( onthefly && (numP == 1)) {
    printf(" ERROR: on-the-fly composition of single process.\n");
    return;
  }

  if ((type < 3) && (numP == 1)) {
    printf(" ERROR: only multi-way composition (///) allows single process.\n");
    return;
  }

  numcomp = Pa_AddNewComposition(name,type,numP,numA);
  if (numcomp == -1) return;

  c = &pa_compositionTable[numcomp];

  /* PROCESSES */

  token = strspn(process," \t");
  sup = (Est_String) strdup(&process[token]);

  token = strcspn(sup," \t");
  s = (Est_String) malloc(token+1);
  strncpy(s,sup,token); /* name of first process */
  s[token] = '\0';

  i = Pa_FindProcess(s);
  free(s);

  if (i == -1) {
    Pa_RemoveComposition(numcomp);
    printf(" ERROR: First process must exist.\n");
    return;
  }

  if (!pa_processTable[i].encoded) {
    Pa_EncodeProcess(pa_processTable[i].name);
    /* if encoding is not possible */
    if (!pa_processTable[i].encoded) return;
    /* make sure that empty process has encoded sort! */
    Pa_EncodeSort(pa_sortTable[pa_processTable[i].sort].name);
  }

  c->tableP[0] = i; /* number of the first process */
  c->sort = pa_processTable[i].sort;
  tau = pa_sortTable[c->sort].table[0].p;

  if ( numP > 1) {

    k = 1;
    while (k < numP) {
      while ((sup[token] == ' ') || (sup[token] == '\t')) token = token + 1;
      sup_next = (Est_String) strdup(&sup[token]);
      free(sup);
      sup = sup_next;

      token = strcspn(sup," \t");
      s = (Est_String) malloc(token+1);
      strncpy(s,sup,token); /* name of next process */
      s[token] = '\0';

      i = Pa_FindProcess(s);
      if (i == -1) {
        printf("\n     WARNING: Process %s does not exist ...",s);
        i = Pa_AddNewProcess(s); /* ADD AN EMPTY PROCESS */
        Pa_FOAStateProcess(&pa_processTable[i],name); /* INITIAL STATE */
        pa_processTable[i].initial = 0;
        pa_processTable[i].sort = c->sort;
      }
      free(s);

      if (c->sort != pa_processTable[i].sort) {
        Pa_RemoveComposition(numcomp);
        printf("\n    COMPOSITION ERROR: Processes have different sorts!\n");
        return;
      }

      if (!pa_processTable[i].encoded) {
        Pa_EncodeProcess(pa_processTable[i].name);
        /* if encoding is not possible */
        if (!pa_processTable[i].encoded) return;
        /* make sure that empty process has encoded sort! */
        Pa_EncodeSort(pa_sortTable[pa_processTable[i].sort].name);
      }

      c->tableP[k++] = i; /* CAREFULLY IF NEW PROCESSES ARE NOT ADDED AT THE END */

    }

    free(sup);
  }

  /* REPORT THE SIZE OF COMPOSING PROCESS */
  /*
  for(i=0; i<numP; i++) {
    printf("\n(COMPOSE %s: %ds, %dt, COPY-BDD: %ds, %dt, ",
      pa_processTable[c->tableP[i]].name,
      pa_processTable[c->tableP[i]].numStates,
      pa_processTable[c->tableP[i]].numTransitions,
      Pa_NumberStateProcess(&pa_processTable[c->tableP[i]]),
      Pa_DecodeProcessTR(&pa_processTable[c->tableP[i]],pa_processTable[c->tableP[i]].d,FALSE)
    );
    if (Bdd_isEqv(pa_sortTable[pa_processTable[c->tableP[i]].sort].table[0].p,bdd_termNull) ||
        Bdd_isEqv(pa_processTable[c->tableP[i]].d,bdd_termNull)) {
      printf("-1v, encoded=%d) ",pa_processTable[c->tableP[i]].encoded);
    } else {
      printf("%dv, encoded=%d) ",
        Pa_DecodeProcessTR(&pa_processTable[c->tableP[i]],
          Bdd_ITE(pa_sortTable[pa_processTable[c->tableP[i]].sort].table[0].p,bdd_termFalse,pa_processTable[c->tableP[i]].d),FALSE),
          pa_processTable[c->tableP[i]].encoded);
    }
    Bdd_SaveFormula("debug_","d",pa_processTable[c->tableP[i]].d);
    Bdd_WriteFunction(stdout,"debug_d");
  }
  */

  /* DO MINIMIZATION OF PROCESSES IF REQUIRED */

  minimize = FALSE;
  for(i=0; i<numP; i++) {

    if (pa_processTable[c->tableP[i]].trace || pa_processTable[c->tableP[i]].weak || pa_processTable[c->tableP[i]].strong) {

      if (!minimize) {
         printf(NEWLINE);
         minimize = TRUE;
      }

      printf("      minimization of process %s ... ",shortname(pa_processTable[c->tableP[i]].name));

      m = -1;

      /* RETURN MINIMAL STRONG-EQUIVALENT PROCESS INSTEAD OF ORIGINAL ONE */
      if (pa_processTable[c->tableP[i]].strong) {
        m = Versis_Minimization(0,0,pa_processTable[c->tableP[i]].name);
      }

      /* RETURN MINIMAL WEAK-EQUIVALENT PROCESS INSTEAD OF ORIGINAL ONE */
      if (pa_processTable[c->tableP[i]].weak) {
        m = Versis_Minimization(1,0,pa_processTable[c->tableP[i]].name);
      }

      /* RETURN MINIMAL DETERMINISTIC TRACE-EQUIVALENT PROCESS INSTEAD OF ORIGINAL ONE */
      if (pa_processTable[c->tableP[i]].trace) {
        m = Versis_Minimization(3,0,pa_processTable[c->tableP[i]].name);
      }

      if (m!=-1) {

        /* REPORT THE SIZE OF ORIGINAL PROCESS */
        printf("(%ds, %llut, %lluv) -> ",
        Pa_NumberStateProcess(&pa_processTable[c->tableP[i]]),
        Pa_DecodeProcessTR(&pa_processTable[c->tableP[i]],pa_processTable[c->tableP[i]].d,FALSE),
        Pa_DecodeProcessTR(&pa_processTable[c->tableP[i]],
          Bdd_ITE(tau,bdd_termFalse,pa_processTable[c->tableP[i]].d),FALSE)
        );

        /* USE THIS IF YOU WANT THAT MINIMIZED PROCESS REPLACES ORIGINAL PROCESS! */
        /*
        sup = strdup(pa_processTable[c->tableP[i]].name);
        Pa_DeleteProcess(c->tableP[i]);
        Pa_RenameProcess(pa_processTable[m].name,sup);
        free(sup);
        */

        /* SET NEW PROCESS */
        c->tableP[i] = m;

        if (c->sort != pa_processTable[m].sort) {
          printf("COMPOSITION ERROR: Minimized process has wrong sort (%s != %s)!\n",
                          pa_sortTable[pa_processTable[m].sort].name,
                          pa_sortTable[c->sort].name);
          return;
        }

        /* REPORT THE SIZE OF MINIMAL PROCESS */
        printf("(%ds, %llut, %lluv) ",
        Pa_NumberStateProcess(&pa_processTable[m]),
        Pa_DecodeProcessTR(&pa_processTable[m],pa_processTable[m].d,FALSE),
        Pa_DecodeProcessTR(&pa_processTable[m],
          Bdd_ITE(tau,bdd_termFalse,pa_processTable[m].d),FALSE)
        );

      }
      printf("OK\n");
    }
  }

  if (minimize) {
       printf("      calculating composition ... ");
  }

  /* ACTIONS */

  if ( numA ) {

    token = strspn(actions," \t");
    sup = (Est_String) strdup(&actions[token]);

    token = strcspn(sup," \t");
    s = (Est_String) malloc(token+1);
    strncpy(s,sup,token); /* name of first action */
    s[token] = '\0';

    i = Pa_FindSortAction(&pa_sortTable[c->sort],s);
    free(s);

    if (i == -1) {
      Pa_RemoveComposition(numcomp);
      printf("    ERROR (composition): Action %%%d does not exist.\n",1);
      return;
    }

    c->tableA[0] = i; /* number of first action */

    k = 1;
    while (k < numA) {
      while ((sup[token] == ' ') || (sup[token] == '\t')) token = token + 1;
      sup_next = (Est_String) strdup(&sup[token]);
      free(sup);
      sup = sup_next;

      token = strcspn(sup," \t");
      s = (Est_String) malloc(token+1);
      strncpy(s,sup,token); /* name of next action */
      s[token] = '\0';

      i = Pa_FindSortAction(&pa_sortTable[c->sort],s);
      free(s);

      if (i == -1) {
        Pa_RemoveComposition(numcomp);
        printf("    ERROR (composition): Action %%%d does not exist.\n",k+1);
        return;
      }
      c->tableA[k++] = i;
    }
    free(sup);
  }

  /* CIRCUITS ONLY */
  /* REMOVE HAZARDS TO GET MORE USEFUL RESULTS FROM THE COMPOSITION */
  if ((numP > 1) && ((type == 6) || (type == 7) || (type == 8))) {
    for(i=0; i<numP; i++) {
      sup1 = Bdd_RelOp(pa_processTable[c->tableP[i]].d,Bdd_NOT(pa_sortTable[c->sort].a0),"#AND Ex xA xS",TRUE); /*?*/
      sup2 = Bdd_RelOp(pa_processTable[c->tableP[i]].d,pa_sortTable[c->sort].a0,"#AND Ex xA xS",TRUE); /*!*/

      supinout = Bdd_ITE(sup1,sup2,bdd_termFalse);
      if (!Bdd_isEqv(supinout,bdd_termFalse)) {
        p = &pa_processTable[c->tableP[i]];
        printf("\n    WARNING: Circuit %s has hazards!",p->name);

        if (!Bdd_isEqv(Bdd_ITE(p->tableS[p->initial].bddR,supinout,bdd_termFalse),bdd_termFalse)) {
          printf("\n    ERROR: Hazards in initial state are not supported!");
        }

        sup = strdup(p->name);
        j = strlen(sup);
        sup = (Est_String) realloc(sup,j+2);
        sup[j]='H';
        sup[j+1]=0;

        /* PARTIALLY COMPOSED PROCESSES ARE NOT SUPPORTED, YET! */
        /* COPY OF NOT-ENCODED OR FULL COMPOSED ENCODED PROCESS IS NOT ENCODED! */

        if ((p->encoded == 1) && (p->numTransitions == 0) && !Bdd_isEqv(p->d,bdd_termFalse)) {
          printf("\n    ERROR: Partially composed processes are not supported!");
        }

        k = Pa_CopyProcess(p->name,sup,NULL,NULL,NULL);

        free(sup);

        p = &pa_processTable[c->tableP[i]]; /* Pa_CopyProcess reallocate process table */
        cp = &pa_processTable[k];
        c->tableP[i] = k;

        for (j=0; j<p->numStates; j++) {
          sup1 = Bdd_ITE(p->tableS[j].bddR,supinout,bdd_termFalse);
          if (!Bdd_isEqv(sup1,bdd_termFalse)) {
            n = strcspn(p->tableS[j].name,"<");
            sup = (Est_String) malloc(n+3);
            memmove(sup,p->tableS[j].name,n);
            sup[n]='#';
            sup[n+1]='H';
            sup[n+2]=0;
            n = Pa_FOAStateProcess(cp,sup);
            free(sup);

            for (k=0; k<p->numTransitions; k++) {
              if ((p->tableT[k].type == 0) && (p->tableT[k].source == j)) {
                Pa_FOATransition(cp,n,p->tableT[k].action,0,p->tableT[k].dest);
                if (p->tableT[k].dest == j) {
                  Pa_FOATransition(cp,n,p->tableT[k].action,0,n);
                }
              }
              if (p->tableT[k].dest == j) {
                Pa_FOATransition(cp,p->tableT[k].source,p->tableT[k].action,p->tableT[k].type,n);
              }

            }
          }
        }

        for (j=0; j<p->numStates; j++) {
          sup1 = Bdd_ITE(p->tableS[j].bddR,supinout,bdd_termFalse);
          if (!Bdd_isEqv(sup1,bdd_termFalse)) {
            for (k=p->numTransitions; k<cp->numTransitions; k++) {
              if (cp->tableT[k].dest == j) {
                n = strcspn(p->tableS[j].name,"<");
                sup = (Est_String) malloc(n+3);
                memmove(sup,p->tableS[j].name,n);
                sup[n]='#';
                sup[n+1]='H';
                sup[n+2]=0;
                n = Pa_FOAStateProcess(cp,sup);
                free(sup);

                Pa_FOATransition(cp,cp->tableT[k].source,cp->tableT[k].action,cp->tableT[k].type,n);
              }
            }
          }
        }

        Pa_EncodeProcess(pa_processTable[c->tableP[i]].name);
        /* if encoding is not possible */
        if (!pa_processTable[c->tableP[i]].encoded) return;
        /* make sure that empty process has encoded sort! */
        Pa_EncodeSort(pa_sortTable[pa_processTable[c->tableP[i]].sort].name);

        printf("\n    WARNING: Replaced with process %s ...",pa_processTable[c->tableP[i]].name);
      }
    }
  }

  /* COLLECT INFORMATION ABOUT ALPHABET OF PROCESSES */
  /* ACTION TAU IS NOT IN THE ALPHABET */
  c->alphabet = (Bdd_Edge *) calloc (numP,sizeof(Bdd_Edge));
  for(k=0; k<numP; k++) {
    sup1 = pa_processTable[c->tableP[k]].d;
    sup1 = Bdd_ITE(tau,bdd_termFalse,sup1);
    c->alphabet[k] = Bdd_RelOpSimple(sup1,"Ex xR xS",TRUE);
  }

  /* CREATE SUPPLEMENTAL STRUCTURES */

  c->di = (Bdd_Edge *) calloc (numP,sizeof(Bdd_Edge));
  c->dij = (Bdd_Edge *) calloc (numP*numP,sizeof(Bdd_Edge));
  c->dinew = (Bdd_Edge *) calloc (numP,sizeof(Bdd_Edge));
  c->dijnew = (Bdd_Edge *) calloc (numP*numP,sizeof(Bdd_Edge));
  c->dicube = (Bdd_Edge *) calloc (numP,sizeof(Bdd_Edge));
  c->dijcube = (Bdd_Edge *) calloc (numP*numP,sizeof(Bdd_Edge));

  /* WITH REGARD TO THE TYPE OF COMPOSITION CALCULATE:                 */
  /* 0 OR 3 OR 6 - actionsBDD CONTAINS RESTRICTED ACTIONS */
  /* 1 OR 4 OR 7 - actionsBDD CONTAINS NON-RESTRICTED ACTIONS AND TAU   */
  /* 2 OR 5 OR 8 - actionsBDD CONTAINS RESTRICTED ACTIONS */

  if ((type == 0) || (type == 3) || (type == 6)) {

    c->actionsBDD = VersisComposeActions(c);

  }

  if ((type == 1) || (type == 4) || (type == 7)) {

    c->actionsBDD = VersisComposeActions(c);
    c->actionsBDD = Bdd_ITE(c->actionsBDD,bdd_termTrue,tau);  /* add action 'TAU' */

  }

  if ((type == 2) || (type == 5) || (type == 8)) {

    c->actionsBDD = VersisComposeActions(c);

  }

  /* CALCULATE INITIAL STATE OF THE COMPOSITION */

  sup1 = bdd_termTrue;
  for(i=0; i<numP; i++) {
    p = &pa_processTable[c->tableP[i]];

    /* DEBUGGING */
    /*
    printf("Versis_Compose (initial state of process %s): %d\n",p->name,p->initial);
    Bdd_FunctionStat(stdout,p->tableS[p->initial].bddR);
    */

    if (p->initial != -1) {
      sup1 = Bdd_ITE(sup1,p->tableS[p->initial].bddR,bdd_termFalse);
    }
  }
  c->initialBDD = sup1;

  /* DEBUGGING */
  /*
  printf("\nThere are %d states in c->initialBDD\n",
          Pa_DecodeCompStates(c,c->initialBDD,FALSE));
  */

  /* SAVE RESULTS */
  for(k=0; k<numP; k++) {
    Bdd_SaveFormula("Alphabet_COMP_",NULL,c->alphabet[k]);
  }
  Bdd_SaveFormula("Actions_COMP_",c->name,c->actionsBDD);
  Bdd_SaveFormula("Initial_COMP_",c->name,c->initialBDD);

  /* PURGE TMP RESULTS */
  /**/
  Bdd_Clean();
  /**/

  /* CALCULATE TRANSITION RELATION OF THE COMPOSITION */

  if (onthefly) {

    /* DEBUGGING */
    /*
    printf("On-the-fly composition!\n");
    */

    c->onthefly = TRUE;
    c->stateBDD = c->initialBDD;
    c->transitionBDD = bdd_termFalse;

  } else {
    int numloops;

    /* DEBUGGING */
    /*
    printf("Regular composition!\n");
    */

    c->onthefly = FALSE;
    c->stateBDD = bdd_termFalse;
    c->transitionBDD = bdd_termFalse;
    Versis_ComposeInit(numcomp);
    numloops = Versis_ComposeStep(numcomp,-1);
    Versis_ComposeFinish(numcomp);

    /* FORCE GARBAGE COLLECTION AFTER COMPOSITION */
    /*
    Bdd_Garbage();
    */

    /* FORCE REORDERING AFTER COMPOSITION - REMOVE ALL FRESH NODES */
    /*
    Bdd_Reorder(c->transitionBDD);
    */

    /* DEBUGGING */
    /*
    printf("\nStatistic after Composition\n");
    printf("Number of loops: %d\n",numloops);
    printf("About system:\n");
    Bdd_SystemStat(stdout);
    */

    /* DEBUGGING */
    /*
    printf("About composition:\n");
    Bdd_FunctionStat(stdout,c->transitionBDD);
    */

    numloops++; /* make compiler happy */

  }

  return;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Versis_ComposeInit(int numcomp)
{
  Pa_Composition *c;
  int i,j,numP;
  Bdd_Edge sup,supin,supout;

  if (numcomp == -1) {
    printf("\n    ERROR (Versis_ComposeInit): Composition does not exist.\n");
    return;
  }

  c = &pa_compositionTable[numcomp];
  numP = c->numProcesses;

  /* REPORT NAMES OF COMPOSING PROCESSES */
  /*
  for(i=0; i<numP; i++) {
    Pa_DecodeProcess(pa_processTable[c->tableP[i]].name,TRUE);
  }
  */

  /* REPORT THE SIZE OF COMPOSING PROCESS */
  /*
  for(i=0; i<numP; i++) {
    printf("\n(COMPOSE %s INIT: PROCESS %s, TABLE: %ds, %dt, BDD: %llut, ",
      c->name,
      pa_processTable[c->tableP[i]].name,
      pa_processTable[c->tableP[i]].numStates,
      pa_processTable[c->tableP[i]].numTransitions,
      Pa_DecodeProcessTR(&pa_processTable[c->tableP[i]],pa_processTable[c->tableP[i]].d,FALSE)
    );
    Bdd_FunctionStat(stdout,pa_processTable[c->tableP[i]].d);
  }
  Bdd_SystemStat(stdout);
  */

  /* REPORT SOME DETAILS OF COMPOSING PROCESS */
  /*
  for(i=0; i<numP; i++) {
    if (Bdd_isEqv(pa_sortTable[pa_processTable[c->tableP[i]].sort].table[0].p,bdd_termNull) ||
        Bdd_isEqv(pa_processTable[c->tableP[i]].d,bdd_termNull)) {
      printf("-1v, encoded=%d)\n",pa_processTable[c->tableP[i]].encoded);
    } else {
      printf("%dv, encoded=%d)\n",
        Pa_DecodeProcessTR(&pa_processTable[c->tableP[i]],
          Bdd_ITE(pa_sortTable[pa_processTable[c->tableP[i]].sort].table[0].p,bdd_termFalse,pa_processTable[c->tableP[i]].d),FALSE),
          pa_processTable[c->tableP[i]].encoded);
    }
  }
  */

  /* FORCE GARBAGE COLLECTION BEFORE COMPOSITION */
  /*
  Bdd_Garbage();
  */

  /* GLOBAL REORDER BEFORE COMPOSITION */
  /*
  Bdd_Reorder(bdd_termNull);
  */

  /* ITERATIVE REORDER BEFORE COMPOSITION */
  /*
  for(i=0; i<numP; i++) {
    Bdd_Reorder(pa_processTable[c->tableP[i]].d);
  }
  */

  /* REPORT THE SIZE OF COMPOSING PROCESS */
  /*
  for(i=0; i<numP; i++) {
    printf("\n(COMPOSE %s AFTER CLEANING/REORDERING: PROCESS %s, TABLE: %ds, %dt, BDD: %llut, ",
      c->name,
      pa_processTable[c->tableP[i]].name,
      pa_processTable[c->tableP[i]].numStates,
      pa_processTable[c->tableP[i]].numTransitions,
      Pa_DecodeProcessTR(&pa_processTable[c->tableP[i]],pa_processTable[c->tableP[i]].d,FALSE)
    );
    Bdd_FunctionStat(stdout,pa_processTable[c->tableP[i]].d);
  }
  Bdd_SystemStat(stdout);
  */

  /* INITIALIZE SUPPLEMENTAL STRUCTURES */
  /* ALSO CHECK FOR CONNECTIONS BETWEEN PROCESSES */
  /* c->dicube AND c->dijcube ARE USED TO HELP FUNCTION Versis_ComposeFinish */
  for(i=0; i<numP; i++) {
    c->di[i] = bdd_termFalse;
    c->dicube[i] = bdd_termFalse;
    if (c->type < 3) {
      /* NOT MULTI-WAY COMPOSITION */
      supout = Bdd_Restrict(c->alphabet[i],Bdd_GetVariable(pa_sortTable[c->sort].a0),TRUE);
    } else {
      /* MULTI-WAY COMPOSITION */
      supout = Bdd_E(c->alphabet[i],Bdd_GetVariable(pa_sortTable[c->sort].a0));
    }
    for(j=0; j<numP; j++) {
      if (c->type < 3) {
        /* NOT MULTI-WAY COMPOSITION */
        supin = Bdd_Restrict(c->alphabet[j],Bdd_GetVariable(pa_sortTable[c->sort].a0), FALSE);
      } else {
        /* MULTI-WAY COMPOSITION */
        supin = Bdd_E(c->alphabet[j],Bdd_GetVariable(pa_sortTable[c->sort].a0));
      }
      sup = Bdd_ITE(supout,supin,bdd_termFalse);
      if (Bdd_isEqv(sup,bdd_termFalse)) {
        c->dij[i * numP + j] = bdd_termNull;
        c->dijcube[i * numP + j] = bdd_termNull;
      } else {
        c->dij[i * numP + j] = bdd_termFalse;
        c->dijcube[i * numP + j] = bdd_termFalse;
      }
    }
  }

  /* INITIALIZE RESULTS */
  c->snew = c->stateBDD = c->initialBDD;
  c->transitionBDD = bdd_termFalse;
}

/**Function****************************************************************
  Synopsis    []
  Description [Calculates next N steps of composition.
               If (N==-1) then loop until final result is obtained.]
  SideEffects [Use Versis_ComposeInit before this.
               c->transitionBDD not calculated.
               Use Versis_ComposeFinish after this.]
  SeeAlso     []
  ************************************************************************/

int
Versis_ComposeStep(int numcomp, int N)
{
  Pa_Composition *c;
  int numP;
  int type;
  int i,j,k;
  Bdd_Edge s_new, s_acc;
  Bdd_Edge sup, sup1, sup2;     /* supplemental functions */
  Bdd_Edge sup2b, sup2bin, sup2bout; /* supplemental functions */
  Bdd_Edge s1, s2, s3;        /* supplemental functions */
  Bdd_Edge new1, new2;        /* supplemental functions */
  Est_String s;
  int loop;

  if (numcomp == -1) {
    printf("\n    ERROR (Versis_ComposeStep): Composition does not exist.\n");
    return -1;
  }

  c = &pa_compositionTable[numcomp];
  numP = c->numProcesses;
  type = c->type;

  /* NOTE: PROCESSES HAVE TO BE ENCODED WITH DIFFERENT STATE-VARIABLES! */
  /* s_new (RRR) CONTAIN NEW STATES FOUND IN PREVIOUSLY STEP OF CALCULATION */
  /* s_acc (RRR) CONTAIN ALL REACHED STATES */

  s_new = c->snew;
  s_acc = c->stateBDD;

  /* c->dinew AND c->dijnew ARE USED TO HELP FUNCTION Versis_ComposeFinish */
  /* not used for multi-way composition */
  for(i=0; i<numP; i++) {
    c->dinew[i] = bdd_termFalse;
    for(j=0; j<numP; j++) {
      if (Bdd_isNull(c->dij[i * numP + j])) {
        c->dijnew[i * numP + j] = bdd_termNull;
      } else {
        c->dijnew[i * numP + j] = bdd_termFalse;
      }
    }
  }

  /* LOOP N TIMES, STOP IF NO NEW STATE IS REACHED */

  loop = 0;
  do {

    loop++;

    /*
    printf("\nLOOP [%d/%d] STARTED\n",loop,N);
    Bdd_SystemStat(stdout);
    */

    /*
    printf("There are %d new states to be processed in loop %d:\n",
           Pa_DecodeCompStates(c,s_new,FALSE),loop);
    */

    /*
    printf("\nNew states to be processed in loop %d:\n",loop);
    Pa_DecodeCompStates(c,s_new,TRUE);
    */

    /* PART ONE */
    /* CALCULATE NEW STATES, WHICH ARE REACHED WITHOUT COMMUNICATION */
    /* BETWEEN PROCESSES AND STORE THEM IN new1 (RRR)                */

    new1 = bdd_termFalse;

    for(i=0; i<numP; i++) {

      /*
      printf("[%d/%d]",i,numP);
      */

      /* COMPUTE ALL TRANSITIONS IN PROCESS I WHICH ARE ALLOWED   */
      /* WITHOUT COMMUNICATION AND STORE THEM IN s1 (RRRAS)       */

      if ((type == 1) || (type == 4) || (type == 7)) {
        s1 = Bdd_ITE(c->actionsBDD,pa_processTable[c->tableP[i]].d,bdd_termFalse);
      }

      if ((type == 0) || (type == 3) || (type == 6) ||
          (type == 2) || (type == 5) || (type == 8))
      {
        s1 = Bdd_ITE(c->actionsBDD,bdd_termFalse,pa_processTable[c->tableP[i]].d);
      }

      /* UPDATE s1 (RRRAS)                                          */
      /* IT NOW CONTAINS ONLY TRANSITIONS STARTING FROM s_new (RRR) */

      /*
      printf("i=%d\n",i,s_new);
      Bdd_FunctionStat(stdout,s_new);
      */
      s1 = Bdd_ITE(s1,s_new,bdd_termFalse);

      /*
      printf("\nCOMPOSE DEBUG (TRANSITIONS WITHOUT COMMUNICATION, NOT TAU):\n");
      Pa_DecodeCompTR(c,Bdd_ITE(pa_sortTable[c->sort].table[0].p,bdd_termFalse,s1),TRUE);
      */

      if (type < 3) {

        /* IF STANDARD CCS COMPOSITION IS COMPUTED THEN:           */
        /* 1. ACCUMULATE TRANSITIONS s1 (RRRAS) IN c->di (RRRASSS) */

        /* c->dinew IS ENOUGH IF Versis_ComposeFinish IS CALLED   */
        /* AFTER EACH CALL OF Versis_ComposeStep, OTHERWISE, USE  */
        /* c->di[i] = Bdd_ITE(c->di[i],bdd_termTrue,s1);          */

        c->dinew[i] = Bdd_ITE(c->dinew[i],bdd_termTrue,s1);

        /* 2. CALCULATE STATES REACHED WITH TRANSITIONS FROM       */
        /*    s1 (RRRASSS) AND STORE THEM IN sup (RRR)             */

        sup = Bdd_NewState1(s1,pa_processTable[c->tableP[i]].name);

        /* 3. ADD STATES FROM sup (RRR) TO new1 (RRR)              */
        new1 = Bdd_ITE(new1,bdd_termTrue,sup);

      } else {

        /* IF MULTI-WAY COMPOSITION WITH ISOCHRONIC FORKS IS COMPUTED THEN: */
        /* 1. CHECK IN OTHER PROCESSES IF THEY CAN PERFORM EQUAL ACTIONS.   */
        /*    STORE COMPLETE TRANSITIONS IN s1 (RRRAS+S)                    */
        /* WARNING: CURRENTLY, INPUT AND OUTPUT ACTIONS MULTISYNC TOGETHER  */
        /* THIS IS OK, BECAUSE IN CIRCUITS OUTPUT ACTIONS ARE UNIQUE        */
        for(j=0; j<numP; j++) {
          if (j != i) {

            if (Bdd_isNull(c->dij[i * numP + j])) {

              /* I AND J CANNOT SYNCHRONISE */
              sup1 = bdd_termFalse;
              sup2b = s1;

            } else {

              sup1 = Bdd_ITE(s1,pa_processTable[c->tableP[j]].d,bdd_termFalse);
              sup2 = Bdd_RelOpSimple(sup1,"Ex xS",TRUE);
              sup2 = Bdd_ITE(sup2,bdd_termFalse,s1);

              /* sup3 = sup1 + sup2 */
              /* sup1 CONTAINS TRANSITIONS, WHERE J IS ALREADY SYNCHRONISED */
              /* sup2 CONTAINS TRANSITIONS, WITH WHICH J CANNOT SYNCHRONISE */

              /* sup2 = sup2a + sup2b */
              /* sup2a ARE THOSE IN ALPHABET OF J                           */
              /* sup2b ARE THOSE NOT IN ALPHABET OF J                       */

              /* TO DO: CHECK IF TAU IS CORRECTLY MANIPULATED! */
              /* TAU IS NOT IN THE ALPHABET */

              /*
              sup2a = Bdd_ITE(c->alphabet[j],sup2,bdd_termFalse);
              */

              sup2b = Bdd_E(c->alphabet[j],Bdd_GetVariable(pa_sortTable[c->sort].a0));
              sup2b = Bdd_ITE(sup2b,bdd_termFalse,sup2);

            }

            /* TRANSITIONS FROM sup2a ARE WRONG - COMPUTATION INTERFERENCE */
            /* TRANSITIONS FROM sup2b ARE ALLOWED, PROCESS J SIMPLY WAIT   */

            /* FUNDAMENTAL-MODE NORMALIZATION */
            /* FOR INPUT ACTIONS, J DOESN'T WAIT IF IT HAS OUTPUT TRANSITION */
            /* STATES MUST NOT HAVE BOTH, INPUT AND OUTPUT ACTIONS */

            if ((type == 6) || (type == 7) || (type == 8)) {

              /*
              printf("\nFUNDAMENTAL-MODE NORMALIZATION - input sup2b:\n");
              Pa_DecodeCompTR(c,sup2b,TRUE);
              */

              sup2bin = Bdd_ITE(pa_sortTable[c->sort].a0,bdd_termFalse,sup2b); /*?*/
              sup2bout = Bdd_ITE(pa_sortTable[c->sort].a0,sup2b,bdd_termFalse); /*!*/

              /* DELETE INPUT TRANISITIONS FROM STATES, WHERE I HAS OUTPUT TRANSITIONS */
              /* HAZARDS MUST BE PREPROCESED ! */
              sup2bin = Bdd_ITE(sup2bout,bdd_termFalse,sup2bin);

              /* DELETE INPUT TRANISITIONS FROM STATES, WHERE J HAS OUTPUT TRANSITIONS */
              /* BUT ONLY FROM THOSE WITHOUT INPUT TRANSITIONS - THERE ARE HAZARDS */
              /* HAZARDS MAY FIRE OR NOT FIRE */

              sup = Bdd_RelOp(pa_processTable[c->tableP[j]].d,pa_sortTable[c->sort].a0,"#AND Ex xA xS",TRUE); /*!*/
              sup2bin = Bdd_ITE(sup,bdd_termFalse,sup2bin);

              /* ALLOWED TRANSITIONS, WHERE ACTION IS NOT IN ALPHABET OF J */
              sup2b = Bdd_ITE(sup2bin,bdd_termTrue,sup2bout);

              /*
              printf("\nFUNDAMENTAL-MODE NORMALIZATION - output sup2b:\n");
              Pa_DecodeCompTR(c,sup2b,TRUE);
              */

            }

            sup2 = Bdd_ITE(sup2b,pa_processTable[c->tableP[j]].stab,
                           bdd_termFalse);
            s1 = Bdd_ITE(sup1,bdd_termTrue,sup2);
        }}

        /* 2. ACCUMULATE TRANSITIONS s1 (RRRASSS) IN c->di (RRRASSS)  */

        c->di[i] = Bdd_ITE(c->di[i],bdd_termTrue,s1);

      } /* else if (type < 3) */
    } /* for (i) */

    /* END OF PART ONE */

    /*
    printf("\nPART ONE [%d/%d] FINISHED\n",loop,N);
    Bdd_SystemStat(stdout);
    */

    /* PART TWO */
    /* CALCULATE NEW STATES, WHICH ARE REACHED WITH COMMUNICATION */
    /* BETWEEN PROCESSES AND STORE THEM IN new2 (RRR)             */

    new2 = bdd_termFalse;

    for(i=0; i<numP; i++) {

      /* FOR ALL J FIND COMMUNICATION BETWEEN PROCESSES I AND J */

      for(j=0; j<numP; j++) {
        if (!Bdd_isNull(c->dij[i * numP + j])) {

          /*
          printf("[%d-%d/%d]",i,j,numP);
          */

          /* COMPUTE TRANSITIONS WITHOUT a0  */
          /* IN PROCESS I WITH OUTPUT ACTIONS s1 (RAS)  */
          /* IN PROCESS J WITH INPUT ACTIONS s2 (RAS)   */
          /* NOT ALL ACTIONS ARE COMPOSED TOGETHER! */

          s1 = Bdd_Restrict(pa_processTable[c->tableP[i]].d,
                          Bdd_GetVariable(pa_sortTable[c->sort].a0), TRUE);

          s2 = Bdd_Restrict(pa_processTable[c->tableP[j]].d,
                          Bdd_GetVariable(pa_sortTable[c->sort].a0), FALSE);

          if ((type == 1) || (type == 4) || (type == 7)) { /* not-restricted actions */
            s1 = Bdd_ITE(c->actionsBDD,bdd_termFalse,s1);
            s2 = Bdd_ITE(c->actionsBDD,bdd_termFalse,s2);
          }

          if ((type == 2) || (type == 5) || (type == 8)) { /* restricted actions */
            s1 = Bdd_ITE(c->actionsBDD,s1,bdd_termFalse);
            s2 = Bdd_ITE(c->actionsBDD,s2,bdd_termFalse);
          }

          if (type < 3) {

            /* IF STANDARD CCS COMPOSITION IS COMPUTED THEN:                  */
            /* 1. COMPOSE s1 (RAS) AND s2 (RAS) TOGETHER AND CREATE s3 (RRSS) */
            /* 2. UPDATE s3 (RRRSS) TO CONTAIN ONLY TRANSITIONS FROM s_new    */

            s3 = Bdd_RelOp(s1,s2,"#AND Ex xA",TRUE);
            s3 = Bdd_ITE(s_new,s3,bdd_termFalse);

            /*
            printf("COMPOSE DEBUG (TRANSITIONS WITH COMMUNICATION):\n");
            Pa_DecodeCompTR(c,Bdd_ITE(Bdd_ITE(s3,s1,bdd_termFalse),s2,bdd_termFalse),TRUE);
            */

            /* 3. ACCUMULATE TRANSITIONS s3 (RRRSS) IN c->dij (RRRSS)  */

            /* c->dijnew IS ENOUGH IF Versis_ComposeFinish IS CALLED   */
            /* AFTER EACH CALL OF Versis_ComposeStep, OTHERWISE, USE   */
            /* c->dij[i * numP + j] = Bdd_ITE(c->dij[i * numP + j],bdd_termTrue,s3); */

            c->dijnew[i * numP + j] = Bdd_ITE(c->dijnew[i * numP + j],
                                              bdd_termTrue, s3); 

            /* 4. CALCULATE STATES REACHED WITH TRANSITIONS FROM   */
            /*    s3 (RRRSS) AND STORE THEM IN sup (RRR)           */

            sup = Bdd_NewState2(s3,
                              pa_processTable[c->tableP[i]].name,
                              pa_processTable[c->tableP[j]].name);

            /* 5. ADD STATES FROM sup (RRR) TO new2 (RRR) */
            new2 = Bdd_ITE(new2,bdd_termTrue,sup);

          } else {

            /* IF MULTI-WAY COMPOSITION IS COMPUTED THEN                        */
            /* 1. COMPOSE s1 (RAS) AND s2 (RAS) TOGETHER AND CREATE s3 (RRASS)  */
            /*    s3 (RRASS) DOES NOT CONTAIN a0                                */

            s3 = Bdd_ITE(s1,s2,bdd_termFalse); /* s1 = output(i), s2 = input(j) */

            /* 2. CHECK IN OTHER PROCESSES IF THEY CAN PERFORM THE SAME ACTION. */
            /*    STORE COMPLETE TRANSITIONS IN s3 (RRRASS)                     */
            /* WARNING: CURRENTLY, INPUT AND OUTPUT ACTIONS BOTH MULTISYNC      */
            /* THIS IS OK, BECAUSE IN CIRCUITS OUTPUT ACTIONS ARE UNIQUE        */

            for(k=0; k<numP; k++) {
              if ((k != i) && (k != j)) {
                if ((!Bdd_isNull(c->dij[i * numP + k]))
                    && (!Bdd_isNull(c->dij[j * numP + k]))) {

                  sup1 = Bdd_ITE(s3,pa_processTable[c->tableP[k]].d,bdd_termFalse);
                  sup2 = Bdd_RelOpSimple(sup1,"Ex xS",TRUE);
                  sup2 = Bdd_E(sup2,Bdd_GetVariable(pa_sortTable[c->sort].a0));

                  s = (Est_String) malloc(255);
                  sprintf(s,"Ex xR<%s>",pa_processTable[c->tableP[k]].name);
                  sup2 = Bdd_RelOpComplex(sup2,s,TRUE);
                  free(s);

                  sup2 = Bdd_ITE(sup2,bdd_termFalse,s3);

                  /* sup3 = sup1 + sup2 */
                  /* sup1 CONTAINS TRANSITIONS, WHERE K CAN SYNCHRONISE          */
                  /* sup2 CONTAINS TRANSITIONS, WITH WHICH K CANNOT SYNCHRONISE  */

                  /* sup2 = sup2a + sup2b */
                  /* sup2a ARE THOSE IN ALPHABET OF J                           */
                  /* sup2b ARE THOSE NOT IN ALPHABET OF J                       */

                  /*
                  sup2a = Bdd_ITE(sup2,c->alphabet[k],bdd_termFalse);
                  */

                  /* TO DO: CHECK IF TAU IS CORRECTLY MANIPULATED! */
                  /* TAU IS NOT IN THE ALPHABET */

                  sup2b = Bdd_E(c->alphabet[k],Bdd_GetVariable(pa_sortTable[c->sort].a0));
                  sup2b = Bdd_ITE(sup2b,bdd_termFalse,sup2);

                  /* TRANSITIONS FROM sup2a ARE WRONG - COMPUTATION INTERFERENCE */
                  /* TRANSITIONS FROM sup2b ARE ALLOWED, PROCESS K SIMPLY WAIT   */

                  sup2b = Bdd_ITE(sup2b,pa_processTable[c->tableP[k]].stab,
                                  bdd_termFalse);
                  s3 = Bdd_ITE(sup1,bdd_termTrue,sup2b);

                } else {

                  s3 = Bdd_ITE(s3,pa_processTable[c->tableP[k]].stab,bdd_termFalse);

                }
              }
            }

            /* 3. REMOVE ACTION VARIABLES FROM s3                      */
            /* 4. s3 (RRRSSS) TO CONTAIN ONLY TRANSITIONS FROM s_new   */

            s3 = Bdd_RelOpSimple(s3,"Ex xA",TRUE);
            s3 = Bdd_ITE(s_new,s3,bdd_termFalse);

            /* 5. ACCUMULATE TRANSITIONS s3 (RRRSS) IN c->dij (RRRSS)  */

            c->dij[i * numP + j] = Bdd_ITE(c->dij[i * numP + j], bdd_termTrue, s3);

          }  /* else if (type < 3) */
        }
      } /* for (j) */
    } /* for (i) */

    /* END OF PART TWO */

    /*
    printf("\nPART TWO [%d/%d] FINISHED\n",loop,N);
    Bdd_SystemStat(stdout);
    */

    /* PART THREE */
    /* GET RESULTS OF ONE STEP */

    if (type < 3) {
      /* NOT MULTI-WAY COMPOSITION */
      /* CALCULATION OF NEW STATES */
      s_new = Bdd_ITE(new1,bdd_termTrue,new2);

    } else {
      /* MULTI-WAY COMPOSITION */
      /* CALCULATION OF NEW STATES */

      new1 = bdd_termFalse;
      for(i=0; i<numP; i++) {

        /* CALCULATE STATES REACHED WITH TRANSITIONS c->di[i] (RRRASSS) */
        /* AND STORE THEM IN sup (RRR)                                  */

        sup = Bdd_NewState(c->di[i]);

        /* ADD STATES FROM sup (RRR) TO new1 (RRR) */
        new1 = Bdd_ITE(new1,bdd_termTrue,sup);

      } /* for (i) */

      new2 = bdd_termFalse;
      for(i=0; i<numP; i++) {
        for(j=0; j<numP; j++) {
          if (i != j) {
            if (!Bdd_isNull(c->dij[i * numP + j])) {

              /* CALCULATE STATES REACHED WITH TRANSITIONS FROM   */
              /* c->dij (RRRSS) AND STORE THEM IN sup (RRR)       */

              sup = Bdd_NewState(c->dij[i * numP + j]);

              /* ADD STATES FROM sup (RRR) TO new2 (RRR) */
              new2 = Bdd_ITE(new2,bdd_termTrue,sup);

            }
          }
        }
      }

      /* UPDATE s_new (RRR) */
      s_new = Bdd_ITE(new1,bdd_termTrue,new2);

    }

    /* END OF PART THREE */

    /*
    printf("There are %d new states stored in new1 (loop %d)\n",
           Pa_DecodeCompStates(c,new1,FALSE),loop);
    printf("There are %d new states stored in new2 (loop %d)\n",
           Pa_DecodeCompStates(c,new2,FALSE),loop);
    */

    /*
    printf("\nNew states found and stored in new1:\n");
    Pa_DecodeCompStates(c,new1,TRUE);
    printf("\nNew states found and stored in new2:\n");
    Pa_DecodeCompStates(c,new2,TRUE);
    */

    /* UPDATE s_new (RRR) */
    s_new = Bdd_ITE(s_acc,bdd_termFalse,s_new);

    /* UPDATE s_acc (RRR) */
    s_acc = Bdd_ITE(s_acc,bdd_termTrue,s_new);

    /*
    printf("\nPART THREE [%d,%d] FINISHED\n",loop,N);
    Bdd_SystemStat(stdout);
    */

  } while(!(Bdd_isEqv(s_new,bdd_termFalse)) && ((N==-1) || (loop<N)));

  /* END OF MAIN LOOP */

  /*
  printf("\nLOOP [%d/%d] FINISHED\n",loop,N);
  */

  /* IF FINAL RESULT IS OBTAINED THEN DECREASE loop TO SIGNAL THIS */
  if (Bdd_isEqv(s_new,bdd_termFalse)) loop--;

  c->stateBDD = s_acc;
  c->snew = s_new;

  return loop;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects [TO DO: when step-by-step composition, use info about
               new states to avoid calculating complete stabs again.]
  SeeAlso     []
  ************************************************************************/

void
Versis_ComposeFinish(int numcomp)
{
  Pa_Composition *c;
  int type;
  int i,j,k,numP;
  Bdd_Edge tau,tr1,tr2;
  Bdd_Edge sup,sup1,sup2;

  if (numcomp == -1) {
    printf("\n    ERROR (Versis_ComposeFinish): Composition does not exist.\n");
    return;
  }

  c = &pa_compositionTable[numcomp];
  numP = c->numProcesses;
  type = c->type;
  tau = pa_sortTable[c->sort].table[0].p;

  /* NODE MANAGEMENT */
  /* on-the-fly model checker should not call cleaning */
  Bdd_SaveTmpFormula(c->snew,(numP)*(numP+1)+1); /* just long enough, we need this in next ComposeStep */
  if (c->onthefly) {
    Bdd_SaveTmpFormula(c->stateBDD,(numP)*(numP+1)+1); /* just long enough, we need this in next ComposeStep */
  } else {
    Bdd_SaveFormula("S_COMP_",c->name,c->stateBDD); /* already a result */
  }

  if (type < 3) {

    /* IF STANDARD CCS COMPOSITION IS COMPUTED      */
    /* dinew AND dijnew WITH ADDED stab MUST BE COLLECTED */

    /* NODE MANAGEMENT */
    /* on-the-fly model checker should not call cleaning */
    for(i=0; i<numP; i++) {
      Bdd_SaveTmpFormula(c->dinew[i],i);
      Bdd_SaveTmpFormula(c->dicube[i],(numP)*(numP+1)+1);
      for(j=0; j<numP; j++) {
        if (!Bdd_isNull(c->dij[i * numP + j])) {
          Bdd_SaveTmpFormula(c->dijnew[i * numP + j],(numP)*(i+1)+j);
          if ((i<j) || Bdd_isNull(c->dijcube[j * numP + i])
                    || Bdd_isEqv(c->dijcube[j * numP + i],bdd_termFalse)) {
            Bdd_SaveTmpFormula(c->dijcube[i * numP + j],(numP)*(numP+1)+1);
          }
        }
      }
    }

    /* ADD stab AND COLLECT dinew */

    tr1 = bdd_termFalse;
    for(i=0; i<numP; i++) {

      /* ALTERNATIVE 1 */
      /**/
      if (Bdd_isEqv(c->dinew[i],bdd_termFalse)) {
        sup = bdd_termFalse;
      } else {
        if (Bdd_isEqv(c->dicube[i],bdd_termFalse)) {
          sup = bdd_termTrue;
          for(k=0; k<numP; k++) {
            if (k != i) {
              sup = Bdd_ITE(sup,pa_processTable[c->tableP[k]].stablist,bdd_termFalse);
            }
          }
          c->dicube[i] = sup;

          /* NODE MANAGEMENT */
          /* we need c->dicube[i] in next ComposeFinish */
          /* on-the-fly model checker should not call cleaning */
          Bdd_SaveTmpFormula(c->dicube[i],(numP)*(numP+1)-i+1);
        }

        sup = Bdd_StabRS(c->dinew[i],c->dicube[i]);
      }
      /**/

      /* ALTERNATIVE 2 */
      /*
      sup = bdd_termTrue;
      for(k=0; k<numP; k++) {
        if (k != i) {
          sup = Bdd_ITE(sup,pa_processTable[c->tableP[k]].stab,bdd_termFalse);
        }
      }
      sup = Bdd_ITE(sup,c->dinew[i],bdd_termFalse);
      */

      /* ALTERNATIVE 3 */
      /*
      sup = c->dinew[i];
      for(k=0; k<numP; k++) {
        if (k != i) {
          sup = Bdd_ITE(sup,pa_processTable[c->tableP[k]].stab,bdd_termFalse);
        }
      }
      */

      tr1 = Bdd_ITE(tr1,bdd_termTrue,sup);

      /* NODE MANAGEMENT */
      Bdd_SaveTmpFormula(tr1,1);
      Bdd_Clean();

    }

    /* NODE MANAGEMENT */
    Bdd_SaveTmpFormula(tr1,(numP)*(numP)); /* just long enough */

    /* ADD stab AND COLLECT dijnew */

    tr2 = bdd_termFalse;
    for(i=0; i<numP; i++) {
      for(j=0; j<numP; j++) {
        if (!Bdd_isNull(c->dijnew[i * numP + j])) {

          /* ALTERNATIVE 1 */
          /**/
          if (Bdd_isEqv(c->dijnew[i * numP + j],bdd_termFalse)) {
            sup = bdd_termFalse;
          } else {
            if (Bdd_isEqv(c->dijcube[i * numP + j],bdd_termFalse)) {
              if (Bdd_isNull(c->dijcube[j * numP + i]) || Bdd_isEqv(c->dijcube[j * numP + i],bdd_termFalse)) {
                sup = bdd_termTrue;
                for(k=0; k<numP; k++) {
                  if ((k != i) && (k != j)) {
                    sup = Bdd_ITE(sup,pa_processTable[c->tableP[k]].stablist,bdd_termFalse);
                  }
                }
                c->dijcube[i * numP + j] = sup;

                /* NODE MANAGEMENT */
                /* we need c->dijcube[i * numP + j] in next ComposeFinish */
                /* on-the-fly model checker should not call cleaning */
                Bdd_SaveTmpFormula(sup,(numP)*(numP-i)-j+1);

              } else {
                c->dijcube[i * numP + j] = c->dijcube[j * numP + i];
              }
            }

            sup = Bdd_StabRS(c->dijnew[i * numP + j],c->dijcube[i * numP + j]);
          }
          /**/

          /* ALTERNATIVE 2 */
          /*
          sup = bdd_termTrue;
          for(k=0; k<numP; k++) {
            if ((k != i) && (k != j)) {
              sup = Bdd_ITE(sup,pa_processTable[c->tableP[k]].stab,bdd_termFalse);
            }
          }
          sup = Bdd_ITE(sup,c->dij[i * numP + j],bdd_termFalse);
          */

          /* ALTERNATIVE 3 */
          /*
          sup = c->dij[i * numP + j];
          for(k=0; k<numP; k++) {
            if ((k != i) && (k != j)) {
              sup = Bdd_ITE(sup,pa_processTable[c->tableP[k]].stab,bdd_termFalse);
            }
          }
          */

          tr2 = Bdd_ITE(tr2,bdd_termTrue,sup);
        }

        /* NODE MANAGEMENT */
        Bdd_SaveTmpFormula(tr2,1);
        Bdd_Clean();

      }
    }
    tr2 = Bdd_ITE(tau,tr2,bdd_termFalse); /* tr2 = 'tau' * tr2 */

  } else {

    /* IF MULTI-WAY COMPOSITION WITH ISOCHRONIC FORKS IS COMPUTED */
    /* di AND dij MUST BE SIMPLY COLLECTED                        */

    /* NODE MANAGEMENT */
    /* on-the-fly model checker should not call cleaning */
    for(i=0; i<numP; i++) {
      Bdd_SaveTmpFormula(c->di[i],(numP)*(numP+1)+1); /* just long enough, we need this in next ComposeStep */
      for(j=0; j<numP; j++) {
        if (!Bdd_isNull(c->dij[i * numP + j])) {
          Bdd_SaveTmpFormula(c->dij[i * numP + j],(numP)*(numP+1)+1); /* just long enough, we need this in next ComposeStep */
        }
      }
    }

    /* COLLECT di */

    tr1 = bdd_termFalse;
    for(i=0; i<numP; i++) {

      tr1 = Bdd_ITE(tr1,bdd_termTrue,c->di[i]);

      /* NODE MANAGEMENT */
      Bdd_SaveTmpFormula(tr1,1);
      Bdd_Clean();

    }

    /* NODE MANAGEMENT */
    Bdd_SaveTmpFormula(tr1,(numP)*(numP)); /* just long enough */

    /* COLLECT dij */

    tr2 = bdd_termFalse;
    for(i=0; i<numP; i++) {
      for(j=0; j<numP; j++) {

        if (!Bdd_isNull(c->dij[i * numP + j])) {
          tr2 = Bdd_ITE(tr2,bdd_termTrue,c->dij[i * numP + j]);
        }

      }

      /* NODE MANAGEMENT */
      Bdd_SaveTmpFormula(tr2,1);
      Bdd_Clean();

    }
    tr2 = Bdd_ITE(tau,tr2,bdd_termFalse); /* tr2 = 'tau' * tr2 */

  } /* if (type) */

  /* VARIANT 1 */
  /*
  sup = Bdd_ITE(tr1,bdd_termTrue,tr2);
  c->transitionBDD = Bdd_ITE(c->transitionBDD,bdd_termTrue,sup);
  */

  /* VARIANT 2 */
  /**/
  c->transitionBDD = Bdd_ITE(c->transitionBDD,bdd_termTrue,tr1);
  c->transitionBDD = Bdd_ITE(c->transitionBDD,bdd_termTrue,tr2);
  /**/

  sup1 = Bdd_RelOpSimple(c->transitionBDD,"Ex xA xS",TRUE);
  sup2 = Bdd_RelOpSimple(c->transitionBDD,"Ex xA xR S2R",TRUE);
  c->deadlockedBDD = Bdd_ITE(sup1,bdd_termFalse,sup2);

  if (c->onthefly) {
    Bdd_SaveTmpFormula(c->deadlockedBDD,1);
    Bdd_SaveTmpFormula(c->transitionBDD,(numP)*(numP+1)+1);
    Bdd_Clean();
  } else {
    Bdd_SaveFormula("SD_COMP_",c->name,c->deadlockedBDD);
    Bdd_SaveFormula("D_COMP_",c->name,c->transitionBDD);
    Bdd_Clean();
  }

  /* DEBUGGING */
  /*
  printf("COMP transitionBDD: %llu\n",Pa_DecodeCompTR(c,c->transitionBDD,FALSE));
  Bdd_FunctionStat(stdout,c->transitionBDD);
  printf("COMP deadlockedBDD\n");
  Bdd_FunctionStat(stdout,c->deadlockedBDD);
  printf("COMP c->snew\n");
  Bdd_FunctionStat(stdout,c->snew);
  printf("COMP c->stateBDD\n");
  Bdd_FunctionStat(stdout,c->stateBDD);
  for(i=0; i<numP; i++) {
    printf("COMP c->dicube[%d]\n",i);
    Bdd_FunctionStat(stdout,c->dicube[i]);
    for(j=0; j<numP; j++) {
      if (!Bdd_isNull(c->dij[i * numP + j])) {
        printf("COMP c->dijcube[%d,%d]\n",i,j);
        Bdd_FunctionStat(stdout,c->dijcube[i * numP + j]);
      }
    }
  }
  */

  /*
  printf("COMP transitionBDD: %llu\n",Pa_DecodeCompTR(c,c->transitionBDD,FALSE));
  printf("COMP transitionBDD visible: %llu\n",Pa_DecodeCompTR(c,Bdd_ITE(tau,bdd_termFalse,c->transitionBDD),FALSE));
  */

}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
wordNumber(Est_String s)
{
  int num;
  int token;
  Est_String sup,sup_next;

  if (!s) return 0;

  num = 0;
  sup = strdup(s);

  token = strspn(sup," \t");
  while (strspn(sup," \t") < strlen(sup)) {
    sup_next = (Est_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    num++;
    token = strcspn(sup," \t");
    sup_next = (Est_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    token = strspn(sup," \t");
  }

  free(sup);
  return num;
}

/**Function****************************************************************
  Synopsis    [Function shortname]
  Description [Remove text after '#']
  SideEffects [This may be different than in other files.]
  SeeAlso     []
  ************************************************************************/

static Est_String
shortname(Est_String name) {
  int i;
  Est_String shortname;

  i = strcspn(name,"#");
  shortname = (char *) malloc(i+1);
  memmove(shortname,name,i);
  shortname[i]=0;
  return shortname;
}

